import unittest
import json
from unittest.mock import Mock, patch

from canalyst_candas.utils.logger import LogFile
from canalyst_candas.utils.requests import (
    get_company_info_from_ticker,
    get_csin_from_ticker,
    get_model_info,
)


class MdsGetInfoFromTickerTests(unittest.TestCase):
    """
    Tests for util methods that hit the {canalyst_candas.settings.CSIN_URL} url using a ticker.

    - get_company_info_from_ticker()
    - get_csin_from_ticker()
    - get_model_info()
    """

    def setUp(self) -> None:
        # From https://mds.canalyst.com/api/equity-model-series/?company_ticker_bloomberg=AAPL+US
        # Note the format of the ticker '{ticker}+{country code}' in the URL
        self.ex_json_str = """
        {
            "count": 1,
            "next": null,
            "previous": null,
            "results": [
                {
                    "csin": "Y8S4N80139",
                    "company": {
                        "name": "Apple Inc.",
                        "self": "https://mds.canalyst.com/api/companies/Y8S4N8/"
                    },
                    "latest_equity_model": {
                        "equity_model_series": {
                            "csin": "Y8S4N80139",
                            "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/"
                        },
                        "model_version": {
                            "name": "FY2021.26",
                            "self": "https://mds.canalyst.com/api/model-versions/Y8S4N80139/periods/FY2021/revisions/26/"
                        },
                        "earnings_update_type": "regular",
                        "published_at": "2021-11-24T22:27:12.252465Z",
                        "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/"
                    },
                    "equity_models": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/"
                }
            ]
        }
        """
        self.json_response = json.loads(self.ex_json_str)
        self.csin = self.json_response["results"][0]["csin"]
        self.latest_mv = self.json_response["results"][0]["latest_equity_model"][
            "model_version"
        ]["name"]
        self.mock_log = Mock(spec=LogFile)
        self.mock_get_request_json_content = patch(
            "canalyst_candas.utils.requests.get_request_json_content"
        ).start()
        self.mock_get_historical_periods = patch(
            "canalyst_candas.utils.requests.get_historical_periods"
        ).start()

    def tearDown(self):
        patch.stopall()

    def test_get_company_info_from_ticker(self):
        self.mock_get_request_json_content.return_value = self.json_response

        result = get_company_info_from_ticker(
            "AAPL US",
            {},
            self.mock_log,
            "mds_host",
        )

        self.assertEqual(result, (self.csin, self.latest_mv))

    def test_get_csin_from_ticker(self):
        self.mock_get_request_json_content.return_value = self.json_response

        result = get_csin_from_ticker(
            "AAPL US",
            {},
            self.mock_log,
            "mds_host",
        )

        self.assertEqual(result, self.csin)

    def test_get_model_info(self):
        historical_periods = [
            "FY2021",
            "Q4-2021",
            "Q3-2021",
            "Q2-2021",
            "Q1-2021",
            "FY2020",
            "Q4-2020",
            "Q3-2020",
            "Q2-2020",
            "Q1-2020",
        ]
        self.mock_get_request_json_content.return_value = self.json_response
        self.mock_get_historical_periods.return_value = historical_periods
        mds_host = "mds_host"

        result = get_model_info(
            "AAPL US",
            {},
            self.mock_log,
            mds_host,
        )

        self.assertEqual(
            result,
            (
                (
                    f"{mds_host}/api/equity-model-series/{self.csin}/equity-models/"
                    f"{self.latest_mv}/historical-data-points/?page_size=500"
                ),
                {
                    "AAPL US": (
                        self.csin,
                        "Apple Inc.",
                        self.latest_mv,
                        historical_periods,
                        "regular",
                        "2021-11-24T22:27:12.252465Z",
                    )
                },
            ),
        )



"""
Connection utilities
"""
from typing import Any

from canalyst_candas.candas import Model, ModelMap, ModelSet, Search
from canalyst_candas.configuration.config import Config, resolve_config


class Connection:
    """
    Connection uses configuration and shares them across the library
    """

    def __init__(self, config: Config = None) -> None:
        self.config = resolve_config(config)

    def get_model(self, *args: Any, **kwargs: Any) -> Model:
        """
        Return a Model with configuration
        """
        return Model(*args, **dict(kwargs, config=self.config))

    def get_model_map(self, *args: Any, **kwargs: Any) -> ModelMap:
        """
        Return a ModelMap with configuration
        """
        return ModelMap(*args, **dict(kwargs, config=self.config))

    def get_model_set(self, *args: Any, **kwargs: Any) -> ModelSet:
        """
        Return a ModelSet with configuration
        """
        return ModelSet(*args, **dict(kwargs, config=self.config))



import io
import string
from typing import List, Optional

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame

from canalyst_candas.configuration.config import Config, create_config
from canalyst_candas.utils.requests import Getter


class Search:
    """
    A class to facilitate Search across the Canalyst Modelverse.

    Parameters
    ----------
    Config object as defined by from canalyst_candas.configuration.config import Config.

    Attributes
    ----------
    ticker_list : list[str]
        list of canalyst tickers
    df_guidance : Pandas DataFrame
        DataFrame of guidance data
    df_search : Pandas DataFrame
        DataFrame of search data

    """

    def __init__(self, config: Config = None) -> None:
        self.config: Config = config or create_config()
        self.df_search: Optional[DataFrame] = None
        self.df_guidance: Optional[DataFrame] = None
        self.ticker_list: Optional[List[str]] = None
        if self.config.s3_access_key_id == "" or self.config.s3_secret_key == "":
            print("Missing S3 keys, please contact Canalyst")

    def score_lists(self, list1, list2, ignore_list):
        """
        A function to perform custom word match score between two lists.

        Parameters
        ----------
        list1: list
        list2: list
        ignore_list: list

        Returns
        ----------
        float
        """

        list1 = [s.replace(",", "") for s in list1]
        list2 = [s.replace(",", "") for s in list2]

        score = 0
        x = len(list2)
        num_matches = 0
        for ii, s in enumerate(list1):

            if s in ignore_list:
                continue

            if s in list2:
                num_matches += 1
                i = list2.index(s) + 1
                z = list1.index(s) + 1
                x = z / i / (ii + 1)
                score = score + x
        if num_matches:
            score = score * num_matches / len(list2)
        return score

    def kpi_match(self, ticker1, ticker2, is_driver=False):
        """
        A function to match KPI between two tickers.

        Parameters
        ----------
        ticker1: str
        ticker2: str
        is_driver: boolean

        Returns
        ----------
        Pandas DataFrame
        """

        from difflib import SequenceMatcher  # we only load this if we need it

        df1 = self.search_time_series(ticker=ticker1, is_driver=is_driver)
        df2 = self.search_time_series(ticker=ticker2, is_driver=is_driver)
        like_list = ["Stats", "Analysis", "Metrics", "Segmented Results", "Revised"]
        list_string = "|".join(like_list).lower()
        df1 = df1.loc[df1["category"].str.lower().str.contains(list_string)]
        df2 = df2.loc[df2["category"].str.lower().str.contains(list_string)]

        list_df = []
        for i, row in df1.iterrows():
            ts = row["time_series_name"]
            td = row["time_series_description"]
            ut = row["unit_type"]
            i_d = row["is_driver"]

            ignore_list = [
                "",
                "significant",
                "growth",
                "effective",
                "other",
                "items",
                "one-time",
                "net",
                "income",
                "y/y",
                "change",
                "consensus",
                "share",
                "price",
                "estimated",
                "dividend",
                "buybacks",
                "adjustments",
                "tax",
                "issuance",
                "%",
                ",",
                "mm",
                " ",
                "-",
                "&",
                "and",
                "#",
                " of",
                "bn",
            ]

            split_description = list(set(td.lower().split(" ")))
            list_out = []
            list_out.append(
                [
                    "ticker_x",
                    "ticker_y",
                    "unit_type",
                    "time_series_description_x",
                    "time_series_name_x",
                    "time_series_description_y",
                    "time_series_name_y",
                    "is_driver_x",
                    "is_driver_y",
                    "word_score",
                    "diff_score",
                ]
            )

            for i2, row2 in df2.iterrows():
                row_description = str(row2["time_series_description"])
                row_name = str(row2["time_series_name"])
                row_unit_type = str(row2["unit_type"])

                if row_unit_type == ut:
                    split_row = list(set(row_description.lower().split(" ")))
                    word_score = self.score_lists(
                        split_description, split_row, ignore_list
                    )
                    diff_score = SequenceMatcher(None, ts, row_name).ratio()
                    if word_score > 0:
                        list_out.append(
                            [
                                ticker1,
                                ticker2,
                                ut,
                                td,
                                ts,
                                row_description,
                                row_name,
                                i_d,
                                row["is_driver"],
                                word_score,
                                diff_score,
                            ]
                        )

            df = pd.DataFrame(list_out)
            df.columns = df.iloc[0]
            df = df[1:]
            df["diff_rank"] = df["diff_score"].rank(pct=True)
            df["word_score_rank"] = df["word_score"].rank(pct=True)
            df["word_plus_diff_score"] = df["word_score"] + df["diff_score"]
            list_df.append(df)
        df = (
            pd.concat(list_df)
            .groupby("time_series_description_x")
            .first()
            .reset_index()
        )
        df = df[
            [
                "ticker_x",
                "ticker_y",
                "unit_type",
                "time_series_name_x",
                "time_series_description_x",
                "time_series_name_y",
                "time_series_description_y",
                "is_driver_x",
                "is_driver_y",
                "word_score",
                "diff_score",
                "word_plus_diff_score",
            ]
        ]

        return df.sort_values("word_plus_diff_score", ascending=False)

    def remove_stopwords(self, sentence):
        """
        A function to remove stopwords from a sentence.

        Parameters
        ----------
        sentence: str

        Returns
        ----------
        str
        """

        my_stopwords = [
            "in",
            "Change",
            "YoY",
            ",",
            "%",
            "(Calculated)",
            "(QoQ)",
            "QoQ",
            "mm",
        ]
        tokens = sentence.split(" ")
        tokens_filtered = [word for word in tokens if not word in my_stopwords]
        tokens_filtered = [
            "".join(c for c in s if c not in string.punctuation)
            for s in tokens_filtered
        ]
        return (" ").join(tokens_filtered)

    def kpi_statistics(self, ticker="", ticker2="", mo_only=False, subsector=""):
        """
        Return the time series correlations inside GIC subsectors.

        Parameters
        -------
        ticker : str
            Ticker with which to filter the available result set.
        ticker2 : str
            Ticker to filter against previous ticker
        mo_only : boolean "False"
            Filter to only common (MO_) names
        subsector : str
            Filter subsector

            Return the rsq between time series for two stocks

        Returns
        -------
        DataFrame

        """

        GET = Getter(config=self.config)
        path_name = f"DATA/rsq.csv"
        df = GET.get_csv_from_s3(path_name)
        df = df.loc[:, ~df.columns.str.contains("^Unnamed")]
        df = df[
            ~df["time_series_description_x"].str.contains("other", na=False, case=False)
        ]
        df = df[
            ~df["time_series_description_y"].str.contains("other", na=False, case=False)
        ]
        df = df[
            ~df["time_series_description_x"].str.contains("tax", na=False, case=False)
        ]
        df = df[
            ~df["time_series_description_y"].str.contains("tax", na=False, case=False)
        ]

        if subsector != "":
            df = df.loc[df["subsector"] == subsector]
            if mo_only == True:
                df1 = df.loc[df["time_series_name_y"].str.startswith("MO")]
                df2 = df.loc[df["time_series_name_x"].str.startswith("MO")]
                df = pd.concat([df1, df2])
            return df.sort_values("rsquared", ascending=False)
        if ticker2 != "":
            df1 = df[(df["ticker_y"] == ticker) & (df["ticker_x"] == ticker2)]
            df2 = df[(df["ticker_x"] == ticker) & (df["ticker_y"] == ticker2)]
        else:
            df1 = df[df["ticker_y"] == ticker]
            df2 = df[df["ticker_x"] == ticker]
        df = pd.concat([df1, df2]).sort_values("rsquared", ascending=False)
        df = df[~df["time_series_description_y"].str.contains("Tax")]
        df = df.loc[:, ~df.columns.str.contains("^Unnamed")]
        if mo_only == True:
            df1 = df.loc[df["time_series_name_y"].str.startswith("MO")]
            df2 = df.loc[df["time_series_name_x"].str.startswith("MO")]
            df = pd.concat([df1, df2])
            if df.shape[0] != 0:
                df = df.sort_values("rsquared", ascending=False)

        return df

    def get_ticker_list(self, ticker="", company_name=""):
        """
        Return the tickers available on the Canalyst system.

        Parameters
        -------
        ticker : str
            Ticker with which to filter the available result set.

            Easy way to check whether we cover a given stock.

        Returns
        -------
        DataFrame

        """
        if self.ticker_list is None:
            GET = Getter(config=self.config)
            path_name = f"DATA/cwl_all.csv"
            df = GET.get_csv_from_s3(path_name)
            df = df.drop(columns=["CSIN"])
            self.ticker_list = df
        else:
            df = self.ticker_list

        if company_name != "":
            df = df.loc[df["File"].str.contains(company_name, case=False)]

        if ticker != "":
            df_ret = df.loc[df["Ticker Alternate 1"] == ticker]
            if df_ret.shape[0] == 0:
                df_ret = df.loc[
                    df["Ticker Alternate 1"].str.contains(ticker, case=False, na=False)
                ]
            df = df_ret
        df = df.loc[:, ~df.columns.str.contains("^Unnamed")]

        return df

    def search_guidance_time_series(
        self,
        ticker="",
        sector="",
        time_series_name="",
        time_series_description="",
        query="",
        most_recent="",
    ):
        """
        Search guidance time series across all available guidance data.

        Parameters
        ----------
            ticker : str or list
                ticker in AXP US format
            sector : str
                searches the Path string
            time_series_name : str
                simple regex match for time_series_name contains the input string
            time_series_description : str
                whole word match and returns number of whole word matches
            most_recent : True or False
                filter to only the most recent for each time series
            query : str
                filter value against a pandas query e.g. Mid > 5

        Returns
        --------
            DataFrame
        """
        if self.df_guidance is None:
            body = Getter(config=self.config).get_zip_csv_from_s3(
                f"DATA/df_guidance.zip"
            )
            self.df_guidance = pd.read_csv(io.StringIO(body), low_memory=False)

        df = self.df_guidance
        df = df.loc[df["Type.1"] == "Estimate"]

        df = df.rename(
            columns={"Item": "time_series_description", "Item Name": "time_series_name"}
        )

        if most_recent == True:
            df = df.sort_values(["ticker", "time_series_name", "Date"], ascending=False)
            df = df.groupby(["ticker", "time_series_name"]).first().reset_index()

        if type(ticker) == list:
            df = df.loc[df["ticker"].isin(ticker)]
        elif ticker != "":
            df = df.loc[df["ticker"] == ticker]

        if sector != "":
            df = df.loc[df["Path"].str.contains(sector, case=True, regex=True)]
        if time_series_name != "":
            df = df.loc[
                df["time_series_name"].str.contains(
                    time_series_name, case=False, regex=True
                )
            ]

        if time_series_description != "":

            ts_lower = time_series_description.lower()
            split_strings = ts_lower.split(" ")
            df["ts_lower"] = df["time_series_description"].str.lower()
            df["search_matches"] = df["ts_lower"].str.count("|".join(split_strings))
            df = df.loc[df["search_matches"] > 0]

        else:
            df["search_matches"] = 1

        df = df[
            [
                "ticker",
                "Path",
                "Filename",
                "time_series_description",
                "time_series_name",
                "Fiscal Period",
                "Low",
                "Mid",
                "High",
                "Type.1",
                "Date",
                "Link",
                "search_matches",
            ]
        ].sort_values(
            ["search_matches", "ticker", "Date", "time_series_name"], ascending=False
        )

        if query != "":
            df = df.query(query)
            df = df.sort_values(["search_matches", "Mid"], ascending=False)
        return df

    def match_time_series(self, ticker="", time_series=""):
        if type(time_series) == "str":
            time_series = [time_series]

        df_list = []
        for item in time_series:
            df = self.search_time_series(ticker=ticker, time_series_description=item)
            df["target"] = item
            df_list.append(df)

        return pd.concat(df_list)

    def search_time_series_excel(
        self,
        ticker="",
        time_series_name="",
        time_series_description="",
        query="",
    ):
        """
        Search time series across all available time series data.
        Optimized for Excel.
        Uses a different (smaller) zip file and reads in chunks.

        Parameters
        ----------
        ticker : str or list[str]
            ticker in AXP US format
        time_series_name : str
            simple regex match for time_series_name contains the input string
        time_series_description : str
            whole word match and returns number of whole word matches
        query : str
            example: "value > 5"

        Returns
        --------
        DataFrame
        """
        pd.options.mode.chained_assignment = None
        print("Building search index...")
        body = Getter(config=self.config).get_zip_csv_from_s3(
            f"DATA/df_search_excel.zip"
        )

        df_list = []
        chunksize = 100000

        for chunk in pd.read_csv(
            io.StringIO(body),
            usecols=["ticker", "time_series_description", "time_series_name", "value"],
            chunksize=chunksize,
        ):
            if type(ticker) == list:
                if len(ticker[0]) > 0:
                    df = chunk.loc[chunk["ticker"].isin(ticker)]
                else:
                    df = chunk
            elif ticker != "":
                print(ticker)
                df = chunk.loc[chunk["ticker"] == ticker]
            else:
                df = chunk

            if time_series_name != "":
                df = df.loc[
                    df["time_series_name"].str.contains(
                        time_series_name, case=False, regex=True
                    )
                ]

            if time_series_description != "":

                ts_lower = time_series_description.lower()

                split_strings = ts_lower.split(" ")
                df["ts_lower"] = df["time_series_description"].str.lower()
                df["search_matches"] = df["ts_lower"].str.count("|".join(split_strings))
                df = df.drop(columns=["ts_lower"])
                df = df.loc[df["search_matches"] > 0]

            else:
                df["search_matches"] = 1

            if df.shape[0] > 0:
                df_list.append(df)
            df = None
            chunk = None

        if len(df_list):
            df = pd.concat(df_list)

        df = df[
            [
                "ticker",
                "time_series_description",
                "time_series_name",
                "value",
                "search_matches",
            ]
        ].sort_values(
            ["search_matches", "ticker", "time_series_description"], ascending=False
        )
        if query != "":
            df = df.query(query)
            df = df.sort_values(["search_matches", "value"], ascending=False)

        return df

    def search_time_series(
        self,
        ticker="",
        sector="",
        time_series_name="",
        time_series_description="",
        category="",
        is_driver="",
        unit_type="",
        mo_only=False,
        exact_match=False,
        period_duration_type="",
        query="",
    ):
        """
        Search time series across all available time series data.

        Parameters
        ----------
            ticker : str or list
                ticker in AXP US format
            time_series_name : str
                simple regex match for time_series_name contains the input string
            time_series_description : str
                whole word match and returns number of whole word matches
            category : str
                category from the path string
            is_driver : str
                is this a driver series
            unit_type : str
                unit type currency, percentage, count, ratio, time
            mo_only : str
                mo names only
            exact_match : str
                only return exact matches
            period_duration_type : str
                fiscal_year or fiscal_quarter
            query : str
                whole word match and returns number of whole word matches

        Returns
        --------
            DataFrame
        """

        pd.options.mode.chained_assignment = None
        if self.df_search is None:
            print("Building search index...")
            body = Getter(config=self.config).get_zip_csv_from_s3(f"DATA/df_search.zip")
            self.df_search = pd.read_csv(
                io.StringIO(body),
                usecols=[
                    "ticker",
                    "CSIN",
                    "Filename",
                    "Path",
                    "publish_date",
                    "update_type",
                    "category",
                    "time_series_description",
                    "time_series_name",
                    "is_driver",
                    "period_duration_type",
                    "unit_type",
                    "value",
                ],
            )

        df = self.df_search

        if type(ticker) == list:
            df = df.loc[df["ticker"].isin(ticker)]
        elif ticker != "":
            df = df.loc[df["ticker"] == ticker]

        if mo_only == True:
            df = df.loc[df["time_series_name"].str.startswith("MO_")]

        if sector != "":
            df = df.loc[df["Path"].str.contains(sector, case=True, regex=True)]

        if category != "":
            df = df.loc[df["category"].str.contains(category, case=True, regex=True)]

        if time_series_name != "":
            if exact_match == True:
                df = df.loc[
                    df["time_series_name"] == time_series_name
                ]  # dont lowercase this
            else:
                df = df.loc[
                    df["time_series_name"].str.contains(
                        time_series_name, case=False, regex=True
                    )
                ]

        if time_series_description != "":

            ts_lower = time_series_description.lower()
            if exact_match == True:
                df = df.loc[df["time_series_description"].str.lower() == ts_lower]
            else:
                split_strings = ts_lower.split(" ")
                df["ts_lower"] = df["time_series_description"].str.lower()
                df["search_matches"] = df["ts_lower"].str.count("|".join(split_strings))
                df = df.drop(columns=["ts_lower"])
                df = df.loc[df["search_matches"] > 0]
        else:
            df["search_matches"] = 1

        if is_driver != "":
            df = df.loc[df["is_driver"] == is_driver]

        if unit_type != "":
            df = df.loc[
                df["unit_type"] == unit_type
            ]  # currency, percentage, count, ratio, time

        if period_duration_type != "":
            df["period_duration_type"] = np.where(
                df["value"].isna(), "fiscal_year", "fiscal_quarter"
            )

        df["period_duration_type"] = np.where(
            df["value"].isna(), "fiscal_year", "fiscal_quarter"
        )

        df = df[
            [
                "ticker",
                "time_series_description",
                "time_series_name",
                "is_driver",
                "category",
                "Path",
                "Filename",
                "period_duration_type",
                "unit_type",
                "value",
                "CSIN",
                "search_matches",
            ]
        ].sort_values(
            ["search_matches", "ticker", "time_series_description"], ascending=False
        )
        if query != "":
            df = df.query(query)
            df = df.sort_values(["search_matches", "value"], ascending=False)

        n = len(pd.unique(df["ticker"]))
        print("No.of.unique tickers: ", n)
        n = len(pd.unique(df["time_series_name"]))
        print("No.of.unique time series: ", n)

        return df



"""
Configs module
"""
import os
import json
from typing import Dict, List, Optional, Any
from pathlib import Path
from pydantic import BaseModel, Field, ValidationError, root_validator
from canalyst_candas import settings
from canalyst_candas.configuration.exceptions import ConfigException


def create_config():
    """
    Create configuration and set to CONFIG for Candas use
    """
    try:
        return resolve_config()
    except ConfigException as config_err:
        config_path = Path.home() / "canalyst"
        config_file = config_path / "keys.json"
        if not os.path.isfile(config_file):
            Path(config_path).mkdir(exist_ok=True)
            config_file_json = {
                "canalyst_api_key": "",
                "s3_access_key_id": "",
                "s3_secret_key": "",
                "fred_key": "",
                "default_dir": "",
                "mds_host": "",
                "wp_host": "",
                "verify_ssl": True,
            }
            try:
                config_file.write_text(json.dumps(config_file_json, indent=2))
                print(
                    "A configuration file has been created for you here -> \n"
                    f"{config_file}."
                )
            except:
                print("Failed to create config file in \n" f"{config_file}.")

        raise config_err


def resolve_config(config: "Config" = None) -> "Config":
    """
    Extract configuration information

    Extracts the information in three ways:
    1. From a passed in Config object
    2. From environment variables
    3. From a configuration file found in the user's home directory
    """
    if config:
        return config

    envar_provider = EnvarProvider()
    envar_config = envar_provider.extract_config()
    if envar_config:
        try:
            return Config(**envar_config)
        except ConfigException as envar_config_err:
            raise ConfigException(
                "Environment variable(s) configuration incorrectly defined"
            ) from envar_config_err

    file_provider = FileProvider()
    file_path = file_provider.extract_config()
    if file_path:
        try:
            return Config.parse_file(file_path)
        except ConfigException as file_config_err:
            raise ConfigException(
                "Configuration file incorrectly defined"
            ) from file_config_err
        except Exception as other_err:
            raise ConfigException from other_err

    raise ConfigException("No configuration was provided")


class Config(BaseModel):
    """
    Configuration validation model
    """

    def __init__(self, config: Dict[str, Any] = {}, **data) -> None:
        """
        Change validation to raise ConfigException
        """
        try:
            super().__init__(**config, **data)
        except ValidationError as error:
            raise ConfigException(error) from None

    canalyst_api_key: str
    s3_access_key_id: str = Field(..., alias="canalyst_s3_id")
    s3_secret_key: str = Field(..., alias="canalyst_s3_key")
    fred_key: str = ""
    default_dir: str = settings.DEFAULT_DIR
    mds_host: str = settings.MDS_HOST
    wp_host: str = settings.WP_HOST
    verify_ssl: bool = True

    class Config:
        allow_population_by_field_name = True

    @root_validator(pre=True)
    def remove_empty_values(cls, values):
        """
        Remove fields with empty values before validation to ensure
        the default values are set for preset fields
        """
        return {k: v for k, v in values.items() if v not in ["", None]}


class ConfigProvider(object):
    """
    A base class for the various configuration providers

    Not to be implemented directly
    """

    def extract_config(self) -> Optional[Dict[str, str]]:
        """
        To be implemented by each provider that inherits from this class
        """
        raise NotImplementedError


class EnvarProvider(ConfigProvider):
    """
    Environment variables configuration provider
    """

    CANALYST_API_KEY = "CANALYST_API_KEY"
    S3_ACCESS_KEY_ID = "S3_ACCESS_KEY_ID"
    S3_SECRET_KEY = "S3_SECRET_KEY"
    FRED_KEY = "FRED_KEY"
    MDS_HOST = "MDS_HOST"
    WP_HOST = "WP_HOST"
    DEFAULT_DIR = "DEFAULT_DIR"
    VERIFY_SSL = "VERIFY_SSL"

    def __init__(
        self, environ: os._Environ = None, mapping: Dict[str, str] = None
    ) -> None:
        self.environ = environ or os.environ
        self._mapping = self._build_mapping(mapping)

    def _build_mapping(self, mapping: Optional[Dict[str, str]]) -> Dict[str, str]:
        """
        Map variable names to environment variables
        """
        if mapping is None:
            return {
                "canalyst_api_key": self.CANALYST_API_KEY,
                "s3_access_key_id": self.S3_ACCESS_KEY_ID,
                "s3_secret_key": self.S3_SECRET_KEY,
                "fred_key": self.FRED_KEY,
                "mds_host": self.MDS_HOST,
                "wp_host": self.WP_HOST,
                "default_dir": self.DEFAULT_DIR,
                "verify_ssl": self.VERIFY_SSL,
            }

        return {
            "canalyst_api_key": mapping.get("canalyst_api_key", self.CANALYST_API_KEY),
            "s3_access_key_id": mapping.get("s3_access_key_id", self.S3_ACCESS_KEY_ID),
            "s3_secret_key": mapping.get("s3_secret_key", self.S3_SECRET_KEY),
            "fred_key": mapping.get("fred_key", self.FRED_KEY),
            "mds_host": mapping.get("mds_host", self.MDS_HOST),
            "wp_host": mapping.get("wp_host", self.WP_HOST),
            "default_dir": mapping.get("default_dir", self.DEFAULT_DIR),
            "verify_ssl": mapping.get("verify_ssl", self.VERIFY_SSL),
        }

    def extract_config(self) -> Optional[Dict[str, str]]:
        """
        Return the configuration information from environment variables
        """
        config = {}

        # only extract set configurations
        for config_var in self._mapping.keys():
            value = self.environ.get(self._mapping[config_var])
            if value is not None:
                config[config_var] = value

        return config


class FileProvider(ConfigProvider):
    """
    Configuration file provider
    """

    def extract_config(self) -> Optional[str]:
        """
        Get configuration file path

        If the config file is not in the default location, search the
        user's home directory for it. If there are multiple matching
        files, use the first one found.
        """
        config_file = f"{Path.home()}/canalyst/keys.json"
        possible_configs: List[Path] = []

        if len(possible_configs) > 1:
            new_line_point = "\n - "
            files = [str(_file) for _file in possible_configs]
            print(
                f"Found multiple configuration files:\n - "
                f"{new_line_point.join(files)}\n\n"
                f"Using {possible_configs[0]}.\n\n"
                "If this not the desired configuration file, remove it. If using a \n"
                "Jupyter Notebook, stop and restart the notebook for the changes to \n"
                "take effect. If using a Python/iPython session, quit the current \n"
                "session and start a new one."
            )

        if os.path.isfile(config_file):
            return config_file



from canalyst_candas.candas.func_music import FuncMusic
from canalyst_candas.candas.chart import Chart
from canalyst_candas.candas.modelset import ModelSet
from canalyst_candas.candas.model import Model, ModelMap
from canalyst_candas.candas.search import Search
from canalyst_candas.candas.sample_model import SampleModel


def help():
    """
    Canalyst Candas help function
    """
    print("Canalyst Candas help")
    print(
        "Please go to https://pypi.org/project/canalyst-candas/#description for installation help"
    )
    print("For support, please contact candas@canalyst.com")
    print("For an API KEY please go to https://app.canalyst.com/u/settings/api-tokens")
    print("For an Excel model download:get_excel_model(ticker, config)")



import unittest
from canalyst_candas import settings

from canalyst_candas.utils.transformations import get_forecast_url


class BaseUtilsTests(unittest.TestCase):
    """
    General test cases for the utils file
    """

    def test_get_forecast_url(self):
        # below URL should actually be valid if you want to manually test
        expected_url = f"{settings.MDS_HOST}/api/equity-model-series/S1TQ5V0161/equity-models/Q3-2021.20/forecast-periods/"
        returned_url = get_forecast_url("S1TQ5V0161", "Q3-2021.20")

        assert expected_url == returned_url



import unittest
import json
from unittest.mock import Mock, patch

from canalyst_candas.utils.logger import LogFile
from canalyst_candas.utils.requests import (
    get_company_info_from_ticker,
    get_csin_from_ticker,
    get_historical_periods,
    get_model_info,
)


class MdsGetHistoricalPeriodsTests(unittest.TestCase):
    """
    Tests for util methods that hit the {canalyst_candas.settings.PERIODS_URL} url using a ticker.
    """

    def setUp(self) -> None:
        # From https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/
        self.ex_json_str = """
        {
            "count": 49,
            "next": null,
            "previous": null,
            "results": [
                {
                    "name": "FY2021",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2020-10-01",
                    "end_date": "2021-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2021/"
                },
                {
                    "name": "Q4-2021",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2021-07-01",
                    "end_date": "2021-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2021/"
                },
                {
                    "name": "Q3-2021",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2021-04-01",
                    "end_date": "2021-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2021/"
                },
                {
                    "name": "Q2-2021",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2021-01-01",
                    "end_date": "2021-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2021/"
                },
                {
                    "name": "Q1-2021",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2020-10-01",
                    "end_date": "2020-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2021/"
                },
                {
                    "name": "FY2020",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2019-10-01",
                    "end_date": "2020-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2020/"
                },
                {
                    "name": "Q4-2020",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2020-07-01",
                    "end_date": "2020-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2020/"
                },
                {
                    "name": "Q3-2020",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2020-04-01",
                    "end_date": "2020-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2020/"
                },
                {
                    "name": "Q2-2020",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2020-01-01",
                    "end_date": "2020-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2020/"
                },
                {
                    "name": "Q1-2020",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2019-10-01",
                    "end_date": "2019-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2020/"
                },
                {
                    "name": "FY2019",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2018-10-01",
                    "end_date": "2019-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2019/"
                },
                {
                    "name": "Q4-2019",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2019-07-01",
                    "end_date": "2019-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2019/"
                },
                {
                    "name": "Q3-2019",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2019-04-01",
                    "end_date": "2019-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2019/"
                },
                {
                    "name": "Q2-2019",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2019-01-01",
                    "end_date": "2019-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2019/"
                },
                {
                    "name": "Q1-2019",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2018-10-01",
                    "end_date": "2018-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2019/"
                },
                {
                    "name": "FY2018",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2017-10-01",
                    "end_date": "2018-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2018/"
                },
                {
                    "name": "Q4-2018",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2018-07-01",
                    "end_date": "2018-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2018/"
                },
                {
                    "name": "Q3-2018",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2018-04-01",
                    "end_date": "2018-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2018/"
                },
                {
                    "name": "Q2-2018",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2018-01-01",
                    "end_date": "2018-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2018/"
                },
                {
                    "name": "Q1-2018",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2017-10-01",
                    "end_date": "2017-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2018/"
                },
                {
                    "name": "FY2017",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2016-10-01",
                    "end_date": "2017-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2017/"
                },
                {
                    "name": "Q4-2017",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2017-07-01",
                    "end_date": "2017-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2017/"
                },
                {
                    "name": "Q3-2017",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2017-04-01",
                    "end_date": "2017-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2017/"
                },
                {
                    "name": "Q2-2017",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2017-01-01",
                    "end_date": "2017-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2017/"
                },
                {
                    "name": "Q1-2017",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2016-10-01",
                    "end_date": "2016-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2017/"
                },
                {
                    "name": "FY2016",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2015-10-01",
                    "end_date": "2016-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2016/"
                },
                {
                    "name": "Q4-2016",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2016-07-01",
                    "end_date": "2016-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2016/"
                },
                {
                    "name": "Q3-2016",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2016-04-01",
                    "end_date": "2016-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2016/"
                },
                {
                    "name": "Q2-2016",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2016-01-01",
                    "end_date": "2016-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2016/"
                },
                {
                    "name": "Q1-2016",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2015-10-01",
                    "end_date": "2015-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2016/"
                },
                {
                    "name": "FY2015",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2014-10-01",
                    "end_date": "2015-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2015/"
                },
                {
                    "name": "Q4-2015",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2015-07-01",
                    "end_date": "2015-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2015/"
                },
                {
                    "name": "Q3-2015",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2015-04-01",
                    "end_date": "2015-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2015/"
                },
                {
                    "name": "Q2-2015",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2015-01-01",
                    "end_date": "2015-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2015/"
                },
                {
                    "name": "Q1-2015",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2014-10-01",
                    "end_date": "2014-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2015/"
                },
                {
                    "name": "FY2014",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2013-10-01",
                    "end_date": "2014-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2014/"
                },
                {
                    "name": "Q4-2014",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2014-07-01",
                    "end_date": "2014-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2014/"
                },
                {
                    "name": "Q3-2014",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2014-04-01",
                    "end_date": "2014-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2014/"
                },
                {
                    "name": "Q2-2014",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2014-01-01",
                    "end_date": "2014-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2014/"
                },
                {
                    "name": "Q1-2014",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2013-10-01",
                    "end_date": "2013-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2014/"
                },
                {
                    "name": "FY2013",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2012-10-01",
                    "end_date": "2013-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2013/"
                },
                {
                    "name": "Q4-2013",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2013-07-01",
                    "end_date": "2013-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q4-2013/"
                },
                {
                    "name": "Q3-2013",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2013-04-01",
                    "end_date": "2013-06-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q3-2013/"
                },
                {
                    "name": "Q2-2013",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2013-01-01",
                    "end_date": "2013-03-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q2-2013/"
                },
                {
                    "name": "Q1-2013",
                    "period_duration_type": "fiscal_quarter",
                    "start_date": "2012-10-01",
                    "end_date": "2012-12-31",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/Q1-2013/"
                },
                {
                    "name": "FY2012",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2011-10-01",
                    "end_date": "2012-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2012/"
                },
                {
                    "name": "FY2011",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2010-10-01",
                    "end_date": "2011-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2011/"
                },
                {
                    "name": "FY2010",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2009-10-01",
                    "end_date": "2010-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2010/"
                },
                {
                    "name": "FY2009",
                    "period_duration_type": "fiscal_year",
                    "start_date": "2008-10-01",
                    "end_date": "2009-09-30",
                    "self": "https://mds.canalyst.com/api/equity-model-series/Y8S4N80139/equity-models/FY2021.26/historical-periods/FY2009/"
                }
            ]
        }
        """
        self.json_response = json.loads(self.ex_json_str)
        self.mock_log = Mock(spec=LogFile)
        self.mock_get_request_json_content = patch(
            "canalyst_candas.utils.requests.get_request_json_content"
        ).start()

    def tearDown(self):
        patch.stopall()

    def test_get_historical_periods(self):
        expected_result = ["FY2021", "Q4-2021", "Q3-2021", "Q2-2021", "Q1-2021"]
        self.mock_get_request_json_content.return_value = self.json_response

        result = get_historical_periods(
            "ticker",
            "model_version",
            {},
            "mds_host",
            self.mock_log,
            num_of_periods_to_fetch=len(
                expected_result
            ),  # only returns the 5 list items
        )

        self.assertEqual(result, expected_result)



"""
GetSampleDrivers tests module
"""
import unittest
from canalyst_candas.utils.requests import get_sample_drivers


class GetSampleDrivers(unittest.TestCase):
    """
    Tests for the get_sample_drivers function
    """

    ticker_options = {
        "disney": "DIS US",
        "netflix": "NFLX US",
        "tesla": "TSLA US",
    }

    def test_return_correct_sample_drivers(self):
        """
        Test that the drivers in the sample dataset are retrieved
        """
        result = get_sample_drivers("DIS US")
        self.assertIsNotNone(result)

    def test_incorrect_sample_driver(self):
        """
        Test that an incorrect ticker is caught
        """
        ticker = "MSFT US"
        with self.assertRaises(FileNotFoundError):
            get_sample_drivers(ticker)



import datetime
from io import StringIO
import io
import sys
from unittest import mock
import unittest
import json
from unittest.mock import Mock, patch

import pandas
from requests.models import Response

from canalyst_candas.settings import BulkDataKeys
from canalyst_candas.utils.logger import LogFile
from canalyst_candas.utils.transformations import _get_mds_bulk_data_url
from canalyst_candas.utils.requests import get_data_set_from_mds


class MdsBulkDataUtilsTests(unittest.TestCase):
    """
    Test cases for the get_data_set_from_mds() method and associated methods
    """

    def setUp(self) -> None:
        self.mock_log = Mock(spec=LogFile)

        self.mock_get_request = patch(
            "canalyst_candas.utils.requests.get_request"
        ).start()

    def tearDown(self):
        sys.stdout = sys.__stdout__
        patch.stopall()

    def test_get_data_set_from_mds_csv_success(self):
        with patch("requests.models.Response") as response_mock:
            some_csv_file = (
                '"header1", "header2", "header3"\n "data1", "data2", "data3"'
            )
            get_csv_response = response_mock
            type(get_csv_response).content = mock.PropertyMock(  # type: ignore
                return_value=bytes(some_csv_file, "utf-8")
            )

            self.mock_get_request.return_value = get_csv_response

            result = get_data_set_from_mds(
                BulkDataKeys.HISTORICAL_DATA,
                "csv",
                "ABCDE12345",
                "Q1-2021.20",
                {},
                self.mock_log,
                "mds_host",
            )

            self.assertEqual(
                result.to_string(), pandas.read_csv(StringIO(some_csv_file)).to_string()
            )

    def test_get_data_set_from_mds_parquet_success(self):
        with patch("requests.models.Response") as response_mock:
            input_df = pandas.DataFrame.from_dict(
                {
                    "period_name": ["FY2020", "Q1-2021", "Q1-2021"],
                    "time_series_name": ["MO_RIS_REV", "MO_MA_EBITDA", "MO_RIS_REV"],
                    "period_start_date": [
                        datetime.date(2020, 1, 1),
                        datetime.date(2021, 1, 1),
                        datetime.date(2021, 1, 1),
                    ],
                    "period_end_date": [
                        datetime.date(2020, 12, 31),
                        datetime.date(2021, 3, 31),
                        datetime.date(2021, 3, 31),
                    ],
                    "value": ["", "0", "3.14"],
                }
            )
            expected_df = pandas.DataFrame.from_dict(
                {
                    "period_name": ["FY2020", "Q1-2021", "Q1-2021"],
                    "time_series_name": ["MO_RIS_REV", "MO_MA_EBITDA", "MO_RIS_REV"],
                    "period_start_date": ["2020-01-01", "2021-01-01", "2021-01-01"],
                    "period_end_date": ["2020-12-31", "2021-03-31", "2021-03-31"],
                    "value": [float("nan"), 0, 3.14],
                }
            )

            get_parquet_response = response_mock

            parquet_df = input_df.to_parquet()
            type(get_parquet_response).content = mock.PropertyMock(  # type: ignore
                return_value=parquet_df
            )

            self.mock_get_request.return_value = get_parquet_response

            result = get_data_set_from_mds(
                BulkDataKeys.HISTORICAL_DATA,
                "parquet",
                "ABCDE12345",
                "Q1-2021.20",
                {},
                self.mock_log,
                "mds_host",
            )

            self.assertEqual(result.to_string(), expected_df.to_string())

    def test_get_data_set_from_mds_failure_model_info_invalid_type(self):
        captured_output = io.StringIO()
        sys.stdout = captured_output

        result = get_data_set_from_mds(
            BulkDataKeys.MODEL_INFO,
            "parquet",
            "ABCDE12345",
            "Q1-2021.20",
            {},
            self.mock_log,
            "mds_host",
        )

        sys.stdout = sys.__stdout__

        self.assertIsNone(result)
        self.assertEqual(
            "model-info is not available in the requested file type: parquet. It is available in the following type(s): ['csv'].\n",
            captured_output.getvalue(),
        )

    def test_get_data_set_from_mds_failure_fcast_invalid_type(self):
        captured_output = io.StringIO()
        sys.stdout = captured_output

        result = get_data_set_from_mds(
            BulkDataKeys.FORECAST_DATA,
            "pickle",
            "ABCDE12345",
            "Q1-2021.20",
            {},
            self.mock_log,
            "mds_host",
        )

        sys.stdout = sys.__stdout__

        self.assertIsNone(result)
        self.assertEqual(
            "forecast-data is not available in the requested file type: pickle. It is available in the following type(s): ['csv', 'parquet'].\n",
            captured_output.getvalue(),
        )

    def test_get_mds_bulk_data_url_csv(self):
        expected = "https://mds.canalyst.com/api/equity-model-series/WZ0R430135/equity-models/FY2021.22/bulk-data/forecast-data.csv"

        result = _get_mds_bulk_data_url(
            BulkDataKeys.FORECAST_DATA,
            "csv",
            "WZ0R430135",
            "FY2021.22",
            "https://mds.canalyst.com",
        )

        self.assertEqual(result, expected)

    def test_get_mds_bulk_data_url_parquet(self):
        expected = "https://mds.canalyst.com/api/equity-model-series/WZ0R430135/equity-models/FY2021.22/bulk-data/name-index.parquet"

        result = _get_mds_bulk_data_url(
            BulkDataKeys.NAME_INDEX,
            "parquet",
            "WZ0R430135",
            "FY2021.22",
            "https://mds.canalyst.com",
        )

        self.assertEqual(result, expected)



"""
Exceptions for candas
"""


class BaseCandasException(Exception):
    """
    Base Candas Exception class
    """


class ScenarioException(BaseCandasException):
    """
    Bad Scenario
    """


class CSINEmptyException(BaseCandasException):
    """
    Empty CSIN
    """


class AWSCredentialException(BaseCandasException):
    """
    Wrong input AWS Credentials
    """


class HTTPClientException(BaseCandasException):
    """
    Wrong input Canalyst API
    """



class FuncMusic(object):
    """
    Helper class for on the fly function generation.

    Used in the ForecastFrame concept where we try to change one param
    across multiple tickers for re-fit in the scenario engine.

    """

    def apply_function(self, value, modifier, argument):
        """
        Apply a specified function to a value

        Parameters
        ----------
        value: int
        modifier: str
        {"add", "subtract", "divide", "multiply","value"}
        argument: int
        """
        self.value = value
        self.modifier = modifier
        self.argument = argument
        method_name = "func_" + str(self.modifier)
        method = getattr(
            self,
            method_name,
            lambda: "Invalid function: use add, subtract, divide, or multiply",
        )
        return method()

    def func_add(self):
        return float(self.value) + float(self.argument)

    def func_divide(self):
        return float(self.value) / float(self.argument)

    def func_multiply(self):
        return float(self.value) * float(self.argument)

    def func_subtract(self):
        return float(self.value) - float(self.argument)



from pathlib import Path

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.image as image
import matplotlib.font_manager as fm

from canalyst_candas.settings import BRAND_CONFIG_DEFAULTS

plt.style.use("fivethirtyeight")


class Chart:
    """
    Charting class to create consistent and branded charts

    Attributes
    ----------
    x_value : pd.Series
        the data for the x-axis
    y_values : pd.DataFrame
        the data for the y-axis
    labels : list
        a list of strings to specify labels of y_value columns respectively (Default: None)
    axis_labels : list
        a list of tuples of strings to specify x,y axis labels (Default: None)
    title : str
        the title of the chart (Default: None)
    plot_styles : list
        list of chart styles for y-value columns (can specify only 1 style to apply to all columns) (Default: ["line"])
    use_subplots : bool
        bool to indicate if subplots should be used (i.e chart columns on the same graph or not) (Default: False)
    subplot_scale_width : list
        list of ints to specify width scaling for subplots (length of list must match # of subplots) (Default: None)
    subplot_scale_height : list
        list of ints to specify height scaling for subplots (length of list must match # of subplots) (Default: None)
    figure_size : tuple
        tuple to define figure size (Default: (12,9))
    vertical_line : str or pd.Series
        str or DataFrame series to define where to place a vertical line on the graph on the x-axis (Default: None)
    marker : str
        str to define marker type (Default: "o")
    markersize : int
        the size of markers (Default: 8)
    markevery : int or float
        int argument will define plotting of every nth marker from the first data point.  (Default: None)
    plot_config : dict
        extra arguments to be passed into the .plot() command. Arguments must be valid for the plot style being used. (Default: None)
    xtick_rotation : int
        specifies the rotation of x-axis tick labels (Default: 90)
    display_charts_horizontal : bool
        specifies if subplots should be displayed horizontally instead of vertically (Default: False)
    subplot_adjustment : list
        list in the form of [left, bottom, right, top] to tune subplot layout. (Default: [0.125, 0.2, 0.9, 0.9])

    Methods
    -------
    show()
        Displays the graph generated
    build_chart(force)
        Builds a matplotlib chart. if force is set to True, chart will be built from scratch regardless if cache is present.
    """

    def __init__(
        self,
        x_value,
        y_values,
        labels=None,
        axis_labels=None,
        title="",
        plot_styles=[
            "line"
        ],  # If only 1 style defined, it will be applied to all columns
        use_subplots=False,
        subplot_scale_width=None,
        subplot_scale_height=None,
        figure_size=(12, 9),
        vertical_line=None,
        marker="o",
        markersize=8,
        markevery=None,
        plot_config={},
        xtick_rotation=90,
        display_charts_horizontal=False,
        subplot_adjustment=[0.125, 0.2, 0.9, 0.9],
        brand_config=None,
        include_logo_watermark=True,
    ):

        self.x_value = x_value
        self.y_values = y_values
        self.labels = labels
        self.axis_labels = axis_labels
        self.title = title
        self.use_subplots = use_subplots
        self.plot_styles = plot_styles
        self.subplot_scale_width = subplot_scale_width
        self.subplot_scale_height = subplot_scale_height
        self.figure_size = figure_size
        self.vertical_line = vertical_line
        self.markevery = markevery
        self.marker = marker
        self.markersize = markersize
        self.plt_plot_style = "fivethirtyeight"
        self.plot_config = plot_config
        self.xtick_rotation = xtick_rotation
        self.display_charts_horizontal = display_charts_horizontal
        self.subplot_adjustment = subplot_adjustment
        self.include_logo_watermark = include_logo_watermark

        self.brand_config = brand_config or BRAND_CONFIG_DEFAULTS

        self._validate_brand_config()
        self._validate_args()

        self.main_fpath = Path(self.brand_config["title_font_path"])
        self.secondary_fpath = Path(self.brand_config["body_font_path"])

        plt.rc("figure", figsize=self.figure_size)
        self.fig, self.axs = self.build_chart()

    def _validate_brand_config(self):
        try:
            self.brand_config["title_font_path"]
        except KeyError:
            raise KeyError(
                "Title font path not found in chart brand configuration. Please reset or update chart brand configs"
            )

        try:
            self.brand_config["chart_bg_color"]
        except KeyError:
            raise KeyError(
                "Chart background color not found in chart brand configuration. Please reset or update chart brand configs"
            )

        try:
            self.brand_config["body_font_path"]
        except KeyError:
            raise KeyError(
                "Body font path not found in chart brand configuration. Please reset or update chart brand configs"
            )

        try:
            self.brand_config["logo_path"]
        except KeyError:
            raise KeyError(
                "Logo path not found in configuration. Please reset or update configs."
            )

        try:
            self.brand_config["figure_bg_color"]
        except KeyError:
            raise KeyError(
                "Figure background color not found in configuration. Please reset or update configs."
            )

        try:
            self.brand_config["font_color"]
        except KeyError:
            raise KeyError(
                "Font color not found in configuration. Please reset or update configs."
            )

        try:
            self.brand_config["vertical_line_color"]
        except KeyError:
            raise KeyError(
                "Vertical line color not found in configuration. Please reset or update configs."
            )

        try:
            self.brand_config["chart_plot_colors"]
        except KeyError:
            raise KeyError(
                "First chart plot color not found in configuration. Please reset or update configs."
            )

        try:
            self.brand_config["ax_spine_color"]
        except KeyError:
            raise KeyError(
                "Ax spine color not found in configuration. Please reset or update configs."
            )

    def _validate_args(self):
        """
        Validate arguments used in initialization
        """
        if not type(self.plot_styles) == list:
            print(
                "Plot styles must be specified in the form of a list. Please re-instantiate the "
                'object with plot style in the form of "["plot_style_here"]". Available options '
                'are "line", "bar", "barh", "scatter".'
            )
        if not type(self.labels) == list:
            print(
                "Labels must be specified in the form of a list. Please re-instantiate the object"
                'with labels in the form of "["Label", "Label"]" depending on how many columns'
                "are being passed in."
            )

        if isinstance(self.y_values, pd.DataFrame):
            if type(self.labels) == list and not len(self.labels) == len(
                self.y_values.columns
            ):
                print("Warning: Labels have not been appointed for all columns.")

        if isinstance(self.y_values, pd.Series):
            if type(self.plot_styles) == list and len(self.plot_styles) > 1:
                print(
                    "Only 1 plot style needed for the specified y data. Only the first value "
                    "in the list will be used."
                )
        elif isinstance(self.y_values, pd.DataFrame):
            columns = len(self.y_values.columns)
            # If user has specified only 1 plot style for multiple columns, extend it to all columns
            if type(self.plot_styles) == list and len(self.plot_styles) == 1:
                self.plot_styles = self.plot_styles * columns
            elif (
                not type(self.plot_styles) == list
                or not len(self.plot_styles) == columns
            ):
                print(
                    "Please explicitly specify the plot style of all Y value columns in a list "
                    "or 1 style to be applied for all."
                )

    def _add_watermark(self, ax):
        """
        Adds a Canalyst Logo to the top left corner of the graph
        """

        logo_path = self.brand_config["logo_path"]

        logo = image.imread(logo_path)
        starting_x, starting_y, width, height = ax.get_position().bounds

        # Units seem to be in inches:
        # https://matplotlib.org/devdocs/gallery/subplots_axes_and_figures/figure_size_units.html#:~:text=The%20native%20figure%20size%20unit,units%20like%20centimeters%20or%20pixels.
        logo_width = 0.06
        logo_height = 0.06

        # Change the numbers in this array to position your image [left, bottom, width, height])
        ax = plt.axes(
            [
                starting_x + 0.01,
                starting_y + height - logo_height,
                logo_width,
                logo_height,
            ],
            frameon=True,
        )
        ax.imshow(logo)
        ax.axis("off")

    def setup_basic_chart_options(self, fig):
        """
        Sets up figure with basic chart options
        """
        plt.style.use(self.plt_plot_style)

        fig_bg_color = self.brand_config["figure_bg_color"]

        fig.patch.set_facecolor(fig_bg_color)
        # Adjust subplot chart display positions
        left, bottom, right, top = self.subplot_adjustment
        plt.subplots_adjust(left, bottom, right, top)

        font_color = self.brand_config["font_color"]

        mpl.rcParams["text.color"] = font_color
        mpl.rcParams["axes.labelcolor"] = font_color
        mpl.rcParams["xtick.color"] = font_color
        mpl.rcParams["ytick.color"] = font_color

    def build_chart(self, force=False):
        """
        Builds a matplotlib chart. Returns figure and ax(s).
        """

        # Avoid rebuilding if not required
        try:
            if self.fig and self.axs and not force:
                print(
                    f"Skipping re-build as the chart is already built. "
                    f"Please provide a `force=True` argument to force the rebuilding of the chart"
                )
                return self.fig, self.axs
        except AttributeError:
            pass

        if isinstance(self.y_values, pd.Series):
            fig, ax = self._get_chart_without_subplots(is_series=True)
        elif isinstance(self.y_values, pd.DataFrame):
            if not self.use_subplots:
                fig, ax = self._get_chart_without_subplots(is_series=False)
            else:
                # Use subplots
                fig, ax = self._get_chart_with_subplots()

        # If a vertical line is specified, and there are subplots: apply it on all subplots
        # If not, apply it to just the one graph.
        vertical_line_color = self.brand_config["vertical_line_color"]

        if self.vertical_line:
            if type(ax) == np.ndarray:
                for index, item in enumerate(ax):
                    ax[index].axvline(
                        x=self.vertical_line,
                        color=vertical_line_color,
                    )
            else:
                ax.axvline(
                    x=self.vertical_line,
                    color=vertical_line_color,
                    linewidth=2,
                )

        return fig, ax

    def _get_axis_label(self, index):
        """
        Return axis label at position specified.

        Returns None if class not initialized with axis_labels.
        """
        if self.axis_labels is not None:
            try:
                return self.axis_labels[index]
            except IndexError:
                pass

        return None

    def _get_label(self, index):
        """
        Return label at position specified.

        Returns None if class not initialized with labels.
        """
        if self.labels is not None:
            try:
                return self.labels[index]
            except IndexError:
                pass
        return None

    def _get_chart_without_subplots(self, is_series=True):
        """
        Returns figure and ax for charts without subplots for series or DataFrames
        """
        fig, ax = plt.subplots()
        first_element = 0

        font_color = self.brand_config["font_color"]

        # Get and set custom font
        fig.suptitle(
            self.title,
            font=self.main_fpath,
            size="x-large",
            weight="bold",
            y=0.96,
            color=font_color,
        )

        self.setup_basic_chart_options(fig)

        label = self._get_label(first_element)
        axis_label = self._get_axis_label(first_element)

        first_chart_plot_color = self.brand_config["chart_plot_colors"][0]

        # If series is specified, we only have 1 y-value to graph
        if is_series:
            self._set_graph_ax(
                ax,
                self.plot_styles[0],
                self.x_value,
                self.y_values,
                label,
                axis_label,
                first_chart_plot_color,
                self.marker,
                self.markersize,
                self.markevery,
            )
        else:
            # Want to graph multiple y-values on the same graph
            color_count = len(self.brand_config["chart_plot_colors"])
            for index, column in enumerate(self.y_values):
                label = self._get_label(index)
                self._set_graph_ax(
                    ax,
                    self.plot_styles[index],
                    self.x_value,
                    self.y_values[column],
                    label,
                    axis_label,
                    self.brand_config["chart_plot_colors"][
                        index % color_count
                    ],  # Use index modded by length of color_options to select unique colors, up to color_count, then colors repeat.
                    self.marker,
                    self.markersize,
                    self.markevery,
                )

        self._format_graph(ax, fig)
        return fig, ax

    def _get_scale_arguments(self):
        """
        Return width, height scale arguments, if any
        """
        scale_args = {}

        if self.subplot_scale_width and self.subplot_scale_height:
            scale_args = {
                "gridspec_kw": {
                    "width_ratios": self.subplot_scale_width,
                    "height_ratios": self.subplot_scale_height,
                }
            }
        elif self.subplot_scale_width:
            scale_args = {"gridspec_kw": {"width_ratios": self.subplot_scale_width}}
        elif self.subplot_scale_height:
            scale_args = {"gridspec_kw": {"height_ratios": self.subplot_scale_height}}

        return scale_args

    def _get_chart_with_subplots(self):
        """
        Returns figure and ax for charts with subplots for DataFrames
        """
        color_count = len(self.brand_config["chart_plot_colors"])
        columns = len(self.y_values.columns)
        scale_args = self._get_scale_arguments()

        # If specified, display subplots horizontally
        if columns % 2 == 0 and self.display_charts_horizontal:
            row = int(columns / 2)
            row_column = {"nrows": row, "ncols": columns}
        else:
            row_column = {"nrows": columns}

        fig, axs = plt.subplots(
            **row_column,
            **scale_args,
            sharex=True,
        )

        font_color = self.brand_config["font_color"]

        # Set custom Font and figure title
        fig.suptitle(
            self.title,
            font=self.main_fpath,
            size="x-large",
            weight="bold",
            y=0.96,
            color=font_color,
        )

        self.setup_basic_chart_options(fig)

        # Plot each column from the Y-column dataset provided
        for index, column in enumerate(self.y_values):
            label = self._get_label(index)
            axis_label = self._get_axis_label(index)

            self._set_graph_ax(
                axs[index],
                self.plot_styles[index],
                self.x_value,
                self.y_values[column],
                label,
                axis_label,
                self.brand_config["chart_plot_colors"][index % color_count],
                self.marker,
                self.markersize,
                self.markevery,
            )
            self._format_graph(axs[index], fig)

        return fig, axs

    def _set_graph_ax(
        self,
        ax,
        plot_style,
        x_value,
        y_value,
        label,
        axis_labels,
        color,
        marker,
        markersize,
        markevery,
    ):
        """
        Builds a matplotlib chart on the ax provided based on configurations specified.
        """
        font_color = self.brand_config["font_color"]

        # Set up axis labels
        if axis_labels is not None:
            x_label, y_label = axis_labels
            ax.set_xlabel(
                x_label,
                font=self.main_fpath,
                labelpad=15,
                color=font_color,
            )
            ax.set_ylabel(
                y_label,
                font=self.main_fpath,
                labelpad=15,
                color=font_color,
            )

        # Plot the type of bar specified
        if plot_style == "bar":
            ax.bar(
                x_value,
                y_value,
                label=label,
                color=color,
                width=0.4,
                zorder=-1,
                **self.plot_config,
            )
        elif plot_style == "scatter":
            ax.scatter(
                x_value,
                y_value,
                label=label,
                color=color,
                zorder=3,
            )
        elif plot_style == "barh":
            ax.barh(
                x_value,
                y_value,
                label=label,
                color=color,
                zorder=-2,
                **self.plot_config,
            )
        else:  # line graph or catch all
            ax.plot(
                x_value,
                y_value,
                label=label,
                color=color,
                marker=marker,
                markersize=markersize,
                markevery=markevery,
                linewidth=2,
                zorder=1,
                **self.plot_config,
            )

    def _format_graph(self, ax, fig):
        """
        Formats ax's
        """
        ax_spine_color = self.brand_config["ax_spine_color"]

        chart_bg_color = self.brand_config["chart_bg_color"]

        # General grid and chart options
        ax.grid(linewidth=0)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_color(ax_spine_color)
        ax.spines["bottom"].set_color(ax_spine_color)
        ax.set_facecolor(chart_bg_color)

        # Setup Legend
        columns = 1
        if isinstance(self.y_values, pd.DataFrame):
            columns = len(self.y_values.columns)

        # Set custom Font
        font_name = fm.FontProperties(fname=self.secondary_fpath)

        fig.legend(
            loc="lower left",
            bbox_to_anchor=(0.1, 0),
            fancybox=True,
            shadow=True,
            ncol=columns,
            frameon=False,
            prop=font_name,
        )

        font_color = self.brand_config["font_color"]

        # Set up x,y axis tick labels
        config_ticks = {
            "size": 0,
            "labelcolor": font_color,
            "labelsize": 12,
            "pad": 10,
        }
        ax.tick_params(axis="both", **config_ticks)

        if self.include_logo_watermark:
            # Add Canalyst watermark
            self._add_watermark(ax)

        for tick in ax.get_xticklabels():
            tick.set_font(self.secondary_fpath)
        for tick in ax.get_yticklabels():
            tick.set_font(self.secondary_fpath)

        # Set the rotation for tick labels
        for tick in ax.get_xticklabels():
            tick.set_rotation(self.xtick_rotation)

    def show(self):
        """
        Displays Chart
        """
        plt.show()



# from canalyst_candas.configuration.config import Config

# from configuration.config import resolve_config
# from canalyst_candas.utils.transformations import calendar_quarter, df_filter
import yahoo_fin.stock_info as si
import pandas as pd
import numpy as np
import string
import statsmodels.tsa.stattools as ts
from statsmodels.regression.rolling import RollingOLS

import statsmodels.api as ssm
import statsmodels.formula.api as smf

from fredapi import Fred
import pandas as pd
import json

import os
import matplotlib.pyplot as plt

plt.style.use("fivethirtyeight")

from bs4 import BeautifulSoup
import pandas as pd
from datetime import datetime
from urllib.request import urlopen, Request

import warnings


# from nltk.sentiment.vader import SentimentIntensityAnalyzer

web_url = "https://finviz.com/quote.ashx?t="


def get_news_sentiment(ticker_list):
    news_tables = {}

    for tick in ticker_list:
        print(tick)
        url = web_url + tick
        req = Request(url=url, headers={"User-Agent": "Chrome"})
        try:
            response = urlopen(req)
        except:
            continue
        html = BeautifulSoup(response, "html.parser")
        news_table = html.find(id="news-table")
        news_tables[tick] = news_table

    news_list = []

    for file_name, news_table in news_tables.items():
        for i in news_table.findAll("tr"):

            text = i.a.get_text()

            date_scrape = i.td.text.split()

            if len(date_scrape) == 1:
                time = date_scrape[0]

            else:
                date = date_scrape[0]
                time = date_scrape[1]

            tick = file_name.split("_")[0]

            news_list.append([tick, date, time, text])

    vader = SentimentIntensityAnalyzer()  # type: ignore

    columns = ["ticker", "date", "time", "headline"]

    news_df = pd.DataFrame(news_list, columns=columns)

    scores = news_df["headline"].apply(vader.polarity_scores).tolist()

    scores_df = pd.DataFrame(scores)

    news_df = news_df.join(scores_df, rsuffix="_right")

    news_df["date"] = pd.to_datetime(news_df.date).dt.date

    return news_df


def get_betas_wide(betas_list):
    list_df = []
    for item in betas_list:

        df = get_price_data(item)
        list_df.append(df)

    df_long = pd.concat(list_df, axis=0).reset_index()
    df_wide = df_long.pivot_table(
        index=["index"], columns="ticker", values="pct_change"
    )
    return df_wide


def get_parallel_betas_from_list(ticker_list, betas_list):
    df_wide = get_betas_wide(betas_list)
    from joblib import Parallel, delayed
    import multiprocessing

    inputs = range(10)

    def processInput(i):
        return i * i

    num_cores = multiprocessing.cpu_count()
    all_df = Parallel(n_jobs=num_cores)(
        delayed(get_beta_data)(ticker, df_wide, betas_list) for ticker in ticker_list
    )
    df_betas = pd.concat(all_df)
    return df_betas


def get_beta_data(ticker, df_wide, betas_list):
    list_out = []
    print(ticker)
    try:
        df = get_price_data(ticker).reset_index()
    except:
        print("failed")
        return

    dataset = pd.merge(df, df_wide, how="inner", left_on="index", right_on="index")

    beta_lens = [60, 90, 252]  # window for betas
    dict_lens = {}
    for beta_len in beta_lens:
        betas = {}
        ns = {}
        r2s = {}
        for col in dataset.columns:
            if col in betas_list:

                df = dataset[["pct_change", col]].dropna()
                # try:
                beta, r2, n = get_betas(df["pct_change"], df[col], beta_len)
                # except:
                #    print("get_betas error")
                #    return
                betas[col] = beta
                ns[col] = n
                r2s[col] = r2
        dict_lens[beta_len] = [betas, ns, r2s]

        df1 = pd.DataFrame.from_dict(
            dict_lens[beta_len][2], orient="index"
        ).reset_index()

        try:
            df1.columns = ["BETA", ticker]
        except:
            print("df columns error")
            return
        df1 = df1.T
        headers = df1.iloc[0]
        df1 = pd.DataFrame(df1.values[1:], columns=headers)
        df1["ticker"] = ticker
        df1["beta_days"] = beta_len
        list_out.append(df1)
    df = pd.concat(list_out)
    return df


def get_price_data(
    ticker, reset=False, rolling_betas=True, rolling_window=60, index_ticker="^GSPC"
):
    price_data = si.get_data(ticker, start_date="01/01/2009")
    df = pd.DataFrame(price_data)
    df = df[["adjclose"]]
    df["pct_change"] = df.adjclose.pct_change()
    df["log_return"] = np.log(1 + df["pct_change"].astype(float))
    df["ticker"] = ticker
    if reset == True:
        df = df.reset_index()
        df["price_date"] = df["index"]
        df = df.drop(columns=["index"])
    if rolling_betas == True:
        beta = "beta_" + str(rolling_window)
        df["adjclose_plus_1"] = df["adjclose"].shift(-1)
        df["adjclose_plus_5"] = df["adjclose"].shift(-5)
        df["adjclose_plus_10"] = df["adjclose"].shift(-10)

        sp_ticker = index_ticker
        sp_data = si.get_data(sp_ticker, start_date="01/01/2009")
        df_sp = pd.DataFrame(sp_data)
        df_sp = df_sp[["adjclose"]]
        df_sp["pct_change"] = df_sp.adjclose.pct_change()
        df_sp["log_return"] = np.log(1 + df_sp["pct_change"].astype(float))
        df_sp["ticker"] = ticker
        df_sp["mkt_change"] = df_sp["pct_change"]
        df_sp = df_sp.reset_index()
        df_sp["price_date"] = df_sp["index"]
        df_sp = df_sp.drop(columns=["index"])
        df_sp["mktclose"] = df_sp["adjclose"]
        df_sp["mkt_plus_1"] = df_sp["mktclose"].shift(-1)
        df_sp["mkt_plus_5"] = df_sp["mktclose"].shift(-5)
        df_sp["mkt_plus_10"] = df_sp["mktclose"].shift(-10)
        df = pd.merge(
            df,
            df_sp[
                [
                    "ticker",
                    "price_date",
                    "mkt_change",
                    "mktclose",
                    "mkt_plus_1",
                    "mkt_plus_5",
                    "mkt_plus_10",
                ]
            ],
        )

        model = RollingOLS.from_formula(
            "pct_change ~ mkt_change",
            data=df[["pct_change", "mkt_change"]],
            window=rolling_window,
        )
        rres = model.fit()
        df[beta] = rres.params[["mkt_change"]]

        df["plus_1_return"] = df["adjclose_plus_1"] / df["adjclose"] - 1
        df["plus_5_return"] = df["adjclose_plus_5"] / df["adjclose"] - 1
        df["plus_10_return"] = df["adjclose_plus_10"] / df["adjclose"] - 1
        df["mktclose_plus_1_return"] = df["mkt_plus_1"] / df["mktclose"] - 1
        df["mktclose_plus_5_return"] = df["mkt_plus_5"] / df["mktclose"] - 1
        df["mktclose_plus_10_return"] = df["mkt_plus_10"] / df["mktclose"] - 1
        df["expected_1_day_return"] = df[beta] * df["mktclose_plus_1_return"]
        df["expected_5_day_return"] = df[beta] * df["mktclose_plus_5_return"]
        df["expected_10_day_return"] = df[beta] * df["mktclose_plus_10_return"]
        df["alpha_1_day"] = df["plus_1_return"] - df["expected_1_day_return"]
        df["alpha_5_day"] = df["plus_5_return"] - df["expected_5_day_return"]
        df["alpha_10_day"] = df["plus_10_return"] - df["expected_10_day_return"]

    return df


def get_betas(x, y, n=0):
    if n > 0:
        x = x.iloc[
            -n:,
        ]
        y = y.iloc[
            -n:,
        ]
    res = ssm.OLS(y, x).fit()
    beta = res.params[0]
    r2 = res.rsquared
    n = len(x)
    return [beta, r2, n]


def get_fred_data(series_list, config, quarters=True):

    end_date = datetime.today().strftime("%Y-%m-%d")
    FRED = Fred(config.fred_key)  # Fred(config.fred_key)

    df_fred = pd.DataFrame()
    for fred_series in series_list:
        s = pd.DataFrame(
            FRED.get_series(
                fred_series, observation_start="2016-01-01", observation_end=end_date
            )
        ).reset_index()
        s.columns = ["end_date", fred_series]
        s["quarter"] = pd.PeriodIndex(s.end_date, freq="Q")
        # s = calendar_quarter(s, "end_date")
        if quarters == True:
            s = s.groupby("quarter").mean().reset_index()
        if df_fred.shape[0] > 0:
            s = s.drop(columns="end_date")
            df_fred = pd.merge(
                df_fred, s, how="inner", left_on="quarter", right_on="quarter"
            )
        else:
            df_fred = s
    return df_fred


def append_price_data(df):

    df = calendar_quarter(df, "end_date")

    pd.set_option("mode.chained_assignment", None)
    ticker = df.iloc[0]["ticker"]
    stock_ticker = ticker.split(" ")[0]
    df_prices = get_price_data(stock_ticker)
    df_prices = df_prices.reset_index()
    df_prices = df_prices[["pricing_date", "adjclose"]]
    df_prices.columns = ["end_date", "adjclose"]

    df_prices = calendar_quarter(df_prices, "end_date")
    df_prices = df_prices.sort_values("end_date")
    df_prices = df_prices.groupby("end_date_CALENDAR_QUARTER").last().reset_index()

    # fiscal_quarter
    df = df_filter(df, "period_duration_type", ["fiscal_quarter"])  # type: ignore
    dates = list(set(list(df["end_date_CALENDAR_QUARTER"])))
    df_p = df_filter(df_prices, "end_date_CALENDAR_QUARTER", dates)  # type: ignore

    df_p.columns = ["end_date_CALENDAR_QUARTER", "end_date", "value"]
    df_p["ticker"] = ticker
    df_p["period"] = ""
    df_p["period_duration_type"] = "fiscal_quarter"
    df_p["category"] = ""
    df_p["type"] = ""
    df_p["row_header"] = "Stock Price"
    df_p["unit"] = "$"
    df_p = df_p.sort_values("end_date")
    df_1 = df_p[
        [
            "ticker",
            "period",
            "period_duration_type",
            "end_date",
            "category",
            "type",
            "row_header",
            "unit",
            "value",
            "end_date_CALENDAR_QUARTER",
        ]
    ]

    df = pd.concat([df, df_1])

    return df


def calendar_quarter(df, col, datetime=True):
    pd.set_option("mode.chained_assignment", None)
    # translate a date into sort-able and group-able YYYY-mm format.
    df[col] = pd.to_datetime(df[col])

    df[col + "shift"] = df[col] + pd.Timedelta(days=-5)

    df[col + "_CALENDAR_QUARTER"] = df[col + "shift"].dt.to_period("Q")

    df = df.drop(columns=[col + "shift"])
    df[col + "_CALENDAR_QUARTER"] = df[col + "_CALENDAR_QUARTER"].astype(str)

    return df


def get_earnings_dates(ticker):
    df = pd.DataFrame.from_dict(si.get_earnings_history(ticker))
    df["earnings_date"] = df["startdatetime"].astype(str)
    df["earnings_date"] = df["earnings_date"].str.split("T").str[0]
    df["earnings_date"] = pd.to_datetime(df["earnings_date"])
    df = df.dropna()

    return df[
        [
            "ticker",
            "companyshortname",
            "earnings_date",
            "epsestimate",
            "epsactual",
            "epssurprisepct",
        ]
    ]


def regress_dataframe(df, y_col, x_col, y_filter, sort_col="", n=24, z_score=False):
    import datetime as ddt

    if z_score == True:
        df[x_col] = (df[x_col] - df[x_col].mean()) / df[x_col].std(ddof=0)
        df[y_col] = (df[y_col] - df[y_col].mean()) / df[y_col].std(ddof=0)

    if df.dropna().shape[0] == 0:
        print("Not enough datapoints")
        return

    if y_filter != "":
        df = df.loc[df[y_col] == y_filter]

    if sort_col != "":
        df = df.sort_values(sort_col, ascending=False)

    X = df.tail(n)[x_col]
    Y = df.tail(n)[y_col]

    X = ssm.add_constant(X)  # to add constant value in the model
    ols_model = ssm.OLS(
        Y.astype(float), X.astype(float), missing="drop"
    ).fit()  # fitting the model
    predictions = ols_model.summary()  # summary of the model
    return predictions


def get_earnings_and_prices(ticker, index_ticker):
    df_earnings = get_earnings_dates(ticker)
    if df_earnings.shape[0] == 0:
        print("No earnings data for " + ticker)
        return None

    df_prices = get_price_data(
        ticker,
        reset=True,
        rolling_betas=True,
        rolling_window=252,
        index_ticker=index_ticker,
    )
    df = pd.merge(
        df_prices,
        df_earnings,
        how="inner",
        left_on=["ticker", "price_date"],
        right_on=["ticker", "earnings_date"],
    )
    beta = "beta_60"
    if "beta_60" not in df.columns:
        for col in df.columns:
            if "beta" in col:
                beta = col
    df = df[
        [
            "ticker",
            "companyshortname",
            "price_date",
            "earnings_date",
            "epsestimate",
            "epsactual",
            "epssurprisepct",
            "alpha_1_day",
            "alpha_5_day",
            "alpha_10_day",
            beta,
            "plus_1_return",
            "plus_5_return",
            "plus_10_return",
            "mktclose_plus_1_return",
            "mktclose_plus_5_return",
            "mktclose_plus_10_return",
        ]
    ].sort_values("earnings_date", ascending=False)
    df = df.groupby(["ticker", "earnings_date"]).first().reset_index()
    return df


def regress_dataframe_groups(
    df_data, y_name="", return_grouped=True, n_periods=12, regression_warnings=False
):
    regression_warning = {}

    max_time_period = df_data["period_end_date"].max()

    period_names = list(
        df_data.groupby("period_name")
        .first()
        .sort_values("period_end_date")
        .reset_index()
        .tail(n_periods)["period_name"]
    )

    warnings.filterwarnings("ignore", category=RuntimeWarning)
    df_data = df_data.sort_values(["time_series_name", "period_end_date"])
    df_data = df_data[df_data[y_name].notna()]
    df_data = df_data[df_data["value"].notna()]

    def extract_lr(x, y_name):
        test_time_period = x["period_end_date"].max()
        if test_time_period != max_time_period:
            time_series_name = x.iloc[0]["time_series_name"]
            regression_warning[time_series_name] = [
                "Time series ends at " + str(test_time_period)
            ]
            return None

        x = x.loc[x["period_name"].isin(period_names)]
        x = x[x["value"].notna()]
        if x.shape[0] < n_periods and x.shape[0] > 0:
            time_series_name = x.iloc[0]["time_series_name"]
            regression_warning[time_series_name] = [
                "Time series has " + str(x.shape[0]) + " periods"
            ]
        if x.shape[0] == 0:
            return None

        model = smf.ols(formula="value ~ " + y_name, data=x)
        rres = model.fit()

        x["intercept"] = rres.params[0]
        x["n_periods"] = x.shape[0]
        x["rsquared"] = rres.rsquared
        x["slope"] = rres.params[1]

        return pd.DataFrame(x)

    grouped = df_data.groupby("time_series_name")  # group by each time series name
    grouped = grouped.apply(lambda x: extract_lr(x, y_name))

    if regression_warning and regression_warnings:
        print("Regression warning:")
        print(pd.DataFrame.from_dict(regression_warning, orient="index"))

    if return_grouped == True:
        try:
            df_output = (
                grouped.reset_index(drop=True)
                .groupby(
                    [
                        "name_index",
                        "time_series_description",
                        "time_series_name",
                        "category",
                    ]
                )
                .mean()
                .reset_index()
                .sort_values("rsquared", ascending=False)
            )
            # df_output = df_output.loc[df_output["n_periods"] >= n_periods]
            return df_output[
                [
                    "name_index",
                    "time_series_description",
                    "time_series_name",
                    "category",
                    "rsquared",
                    "slope",
                    "intercept",
                    "n_periods",
                ]
            ]
        except:
            print("Regression error")
    else:
        return grouped


def format_borders(plot, colors):
    plot.spines["top"].set_visible(False)
    plot.spines["left"].set_visible(False)
    plot.spines["left"].set_color(colors["grey"])
    plot.spines["bottom"].set_color(colors["grey"])


def plot_prices_against_time_series(
    df_data, df_prices, time_series_name, value_name="value"
):
    ticker = df_data.iloc[0]["ticker"]

    df_filter = df_data.loc[df_data["time_series_name"] == time_series_name]

    time_series_description = df_filter.iloc[0]["time_series_description"]

    df_plot = pd.merge(
        df_prices[["adjclose", "price_date"]],
        df_filter[["time_series_name", "earnings_date", value_name]],
        how="outer",
        left_on="price_date",
        right_on="earnings_date",
    )
    df_plot = df_plot.loc[df_plot["price_date"] > "2020-01-01"]

    df_plot["row_num"] = np.arange(len(df_plot))
    markers_on = df_plot.dropna()

    markers_on = list(markers_on["row_num"])

    colors = {
        "red": "#ff207c",
        "grey": "#42535b",
        "blue": "#00838F",
        "orange": "#ffa320",
        "green": "#00ec8b",
    }
    plt.rc("figure", figsize=(12, 9))

    config_ticks = {"size": 14, "color": colors["grey"], "labelcolor": colors["grey"]}
    config_title = {"size": 18, "color": colors["grey"], "ha": "left", "va": "baseline"}

    fig, axes = plt.subplots(2, 1, gridspec_kw={"height_ratios": [3, 1]}, sharex=True)
    fig.tight_layout(pad=3)
    fig.suptitle(str(ticker) + " price vs " + str(time_series_description), fontsize=16)

    date = df_plot["price_date"]
    close = df_plot["adjclose"]
    vol = df_plot[value_name]

    plot_price = axes[0]
    plot_price.plot(
        date,
        close,
        color=colors["blue"],
        linewidth=2,
        label="Price",
        marker="*",
        ms=10,
        markevery=markers_on,
        markerfacecolor="black",
    )

    plot_vol = axes[1]
    plot_vol.bar(date, vol, width=25, color="darkgrey")
    plot_vol.set_ylabel(time_series_name, fontsize=14)
    plot_vol.yaxis.set_label_position("left")
    plot_vol.yaxis.label.set_color(colors["grey"])

    plot_price.yaxis.tick_right()
    plot_price.tick_params(axis="both", **config_ticks)
    plot_price.set_ylabel("Price (in USD)", fontsize=14)
    plot_price.yaxis.set_label_position("right")
    plot_price.yaxis.label.set_color(colors["grey"])
    plot_price.grid(axis="y", color="gainsboro", linestyle="-", linewidth=1)
    plot_price.set_axisbelow(True)

    format_borders(plot_price, colors)
    format_borders(plot_vol, colors)

    return  # plt.show()


def plot_eps_surprise_data(df, alpha_days="alpha_10_days"):

    ticker = df.iloc[0]["ticker"]

    df["z_score_" + alpha_days] = (df[alpha_days] - df[alpha_days].mean()) / df[
        alpha_days
    ].std(ddof=0)

    # df_z_scaled[column] = (df_z_scaled[column] - df_z_scaled[column].mean()) / df_z_scaled[column].std()

    df["z_score_epssurprisepct"] = (
        df["epssurprisepct"] - df["epssurprisepct"].mean()
    ) / df["epssurprisepct"].std(ddof=0)

    df_plot = df
    i_correlation = round(
        df_plot["z_score_epssurprisepct"].corr(df["z_score_" + alpha_days]), 2
    )

    colors = {
        "red": "#ff207c",
        "grey": "#1d2224",
        "blue": "#00838F",
        "orange": "#ffa320",
        "green": "#00ec8b",
    }
    plt.rc("figure", figsize=(12, 9))

    config_ticks = {"size": 14, "color": colors["grey"], "labelcolor": colors["grey"]}
    config_title = {"size": 18, "color": colors["grey"], "ha": "left", "va": "baseline"}

    # fig, axes = plt#.subplots(, 1, gridspec_kw={"height_ratios": [1, 1]})
    plt.tight_layout(pad=3)

    date = df["price_date"]
    close = df["z_score_" + alpha_days]
    vol = df_plot["z_score_epssurprisepct"]

    plt.scatter(date, close, color=colors["blue"], label="Alpha Z Score")

    plt.scatter(date, vol, color="black", label="Eps Surprise Z Score")

    plt.gca().yaxis.tick_right()
    plt.gca().tick_params(axis="both", **config_ticks)
    plt.gca().set_ylabel("Z Score", fontsize=14)
    plt.gca().yaxis.set_label_position("right")
    plt.grid(axis="y", color="gainsboro", linestyle="-", linewidth=0.5)
    plt.legend(loc="upper left")
    plt.title(
        ticker + " earnings suprise and alpha correlation:" + str(i_correlation),
        fontsize=16,
    )
    format_borders(plt.gca(), colors)

    return


def regress_dataframe_time_series_groups(
    df_data=None,
    y_name="",
    return_grouped=True,
    use_qoq_deltas=False,
    use_yoy_deltas=False,
    category_filter=[],
    category="",
    n_periods=16,
):
    df_data = df_data[
        ~df_data["time_series_name"].str.contains("other", case=False, regex=True)
    ]

    if len(category_filter):
        df_data = df_data.loc[~df_data["category"].isin(category_filter)]

    if type(category) is list:
        df_data = df_data.loc[df_data["category"].isin(category)]
    elif category != "":
        df_data = df_data.loc[df_data["category"] == category]

    if use_qoq_deltas == False:
        df_data = df_data.loc[
            ~df_data["time_series_name"].str.contains("quarter_over_quarter")
        ]

    if use_yoy_deltas == False:
        df_data = df_data.loc[
            ~df_data["time_series_name"].str.contains("year_over_year")
        ]

    ticker = df_data.iloc[0]["ticker"]

    df = regress_dataframe_groups(
        df_data, y_name=y_name, return_grouped=return_grouped, n_periods=n_periods
    )

    if df is None:
        print("Regression error: perhaps not enough datapoints")
        return None

    df["ticker"] = ticker
    df["name_index"] = df["name_index"].astype(int)
    col_name = "ticker"
    first_col = df.pop(col_name)
    df.insert(0, col_name, first_col)

    df = df.loc[df["rsquared"] < 0.98]
    df = df.loc[df["rsquared"] > 0.01]
    df = df.loc[df["n_periods"] >= n_periods]
    return df


def rolling_r_squared(
    df_data, time_series_name, window, y_name="alpha_10_day", plot=False
):
    # df_data = df_data.sort_values('period_end_date',ascending=False)
    category_filter = []
    df = df_data[
        df_data["time_series_name"] == time_series_name
    ]  # [['period_name','value','earnings_date','alpha_10_day']]
    time_series_description = df.iloc[0]["time_series_description"]
    model = RollingOLS.from_formula(
        "value ~ " + str(y_name), data=df[["value", y_name]], window=window
    )
    rres = model.fit()
    df["intercept"] = rres.params[["Intercept"]]
    df["slope"] = rres.params[[y_name]]
    df["rsquared"] = rres.rsquared
    df = df[
        [
            "period_name",
            "earnings_date",
            y_name,
            "value",
            "slope",
            "rsquared",
            "intercept",
        ]
    ].dropna()

    if df["rsquared"].isnull().all():
        print(
            "Regression warning: All rsquareds are null. Time series length is less than the window. Try a shorter window."
        )

    if plot == False:
        return df

    colors = {
        "red": "#ff207c",
        "grey": "#C3C2C3",
        "blue": "#00838F",
        "orange": "#ffa320",
        "green": "#00ec8b",
    }
    plt.rc("figure", figsize=(12, 9))

    fig, (ax1, ax2, ax3) = plt.subplots(nrows=3, sharex=True)
    fig.suptitle(time_series_description, fontsize=16)
    fig.tight_layout(pad=1)

    ax1.plot(df["slope"], label="Regression Slope", color=colors["blue"], linewidth=2)
    ax1.legend(loc="upper left")

    ax2.plot(df["rsquared"], label="R Squared", color=colors["blue"], linewidth=2)
    ax2.legend(loc="upper left")

    ax3.plot(df["value"], label="Time Series Value", color=colors["blue"], linewidth=2)
    ax3.legend(loc="upper left")

    import warnings

    warnings.filterwarnings("ignore")

    ax1.set_xticklabels(df["period_name"])  # annoying matplotlib warning in error

    ax1.get_shared_x_axes().join(ax1, ax2, ax3)

    # ax1.title.set_text(time_series_description,fontsize=14)
    format_borders(ax1, colors)
    format_borders(ax2, colors)
    format_borders(ax3, colors)

    plt.show()


def get_earnings_estimates(ticker):
    return si.get_analysts_info(ticker)["Earnings Estimate"]


def get_revenue_estimates(ticker):
    return si.get_analysts_info(ticker)["Revenue Estimate"]


def get_eps_revisions(ticker):
    return si.get_analysts_info(ticker)["EPS Revisions"]



from pathlib import Path
from enum import Enum
import pkg_resources
import tempfile

EQUITY_MODEL_URL = "api/equity-model-series/{csin}/equity-models/{version}/"  # TODO: possible unused setting: delete if necessary
CSIN_URL = "api/equity-model-series/?company_ticker_bloomberg={ticker}"
PERIODS_URL = (
    "api/equity-model-series/{csin}/equity-models/{version}/historical-periods/"
)
SCENARIO_URL = "api/equity-model-series/{csin}/equity-models/{version}/scenarios/"
SCENARIO_URL_FORECAST = (  # TODO: possible unused setting: delete if necessary
    "api/equity-model-series/{csin}/equity-models/"
    "{version}/scenarios/{scenario_id}/forecast-periods/"
)
BULK_DATA_URL = "api/equity-model-series/{csin}/equity-models/{version}/bulk-data/{bulk_data_key}.{file_type}"

HISTORICAL_PERIODS_TO_FETCH = 36

BRAND_CONFIG_DEFAULTS = {
    "chart_plot_colors": [
        "#E2E3E3",
        "#FFD200",
        "#367ADD",
        "#8CE0FD",
        "#E52B26",
        "#F8623F",
        "#00838F",
        "#30C08F",
        "#8EA1FF",
    ],
    "figure_bg_color": "#FFFFFF",
    "vertical_line_color": "#E2E3E3",
    "ax_spine_color": "#C3C2C3",
    "title_font_path": pkg_resources.resource_filename(
        __name__, "fonts/Barlow-SemiBold.ttf"
    ),
    "body_font_path": pkg_resources.resource_filename(
        __name__, "fonts/Roboto-Regular.ttf"
    ),
    "chart_bg_color": "#161B21",
    "font_color": "#000",
    "logo_path": pkg_resources.resource_filename(__name__, "images/logo.png"),
}


class BulkDataKeys(Enum):
    FORECAST_DATA = "forecast-data"
    HISTORICAL_DATA = "historical-data"
    NAME_INDEX = "name-index"
    MODEL_INFO = "model-info"


class BulkDataFileType(Enum):
    CSV = "csv"
    PARQUET = "parquet"


# Setting it up as a local dict instead of based on the /bulk-data endpoint results
# to allow for faster generation (less queries to the  MDS),
# at the cost of duplication across the MDS & Candas.
BULK_DATA_KEY_FILE_TYPES = {
    BulkDataKeys.FORECAST_DATA: [
        BulkDataFileType.CSV.value,
        BulkDataFileType.PARQUET.value,
    ],
    BulkDataKeys.HISTORICAL_DATA: [
        BulkDataFileType.CSV.value,
        BulkDataFileType.PARQUET.value,
    ],
    BulkDataKeys.NAME_INDEX: [
        BulkDataFileType.CSV.value,
        BulkDataFileType.PARQUET.value,
    ],
    BulkDataKeys.MODEL_INFO: [
        # model info _only_ has CSV due to incompatibility with parquet
        BulkDataFileType.CSV.value
    ],
}

SAMPLE_MODEL_PATH = {
    "drivers_path": pkg_resources.resource_filename(__name__, "sample_data/drivers"),
    "models_path": pkg_resources.resource_filename(__name__, "sample_data/models"),
}

DEFAULT_DIR = tempfile.gettempdir()
ROOT_DIR = Path().resolve()
MDS_HOST = "https://mds.canalyst.com"
WP_HOST = "https://app.canalyst.com"
VERIFY_SSL = True



"""
a class of multiple models

NOTE: prevent the possibility of having circular imports
by avoiding direct dependency on other core classes
"""

from functools import reduce
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import matplotlib.pyplot as plt
from pandas.core.frame import DataFrame

from canalyst_candas.datafunctions import DataFunctions
from canalyst_candas.candas.model import Model
from canalyst_candas.candas.chart import Chart
from canalyst_candas.candas.func_music import FuncMusic
from canalyst_candas.configuration.config import Config, create_config
from canalyst_candas.exceptions import ScenarioException
from canalyst_candas.utils.logger import LogFile
from canalyst_candas.utils.transformations import (
    df_filter,
    filter_dataset,
)
from canalyst_candas.utils.requests import (
    get_api_headers,
    send_scenario,
)

plt.style.use("fivethirtyeight")


class ModelSet:
    """
    Generate a dictionary of Model objects, provide ease-of-use functions across the dictionary.

    Parameters
    ----------
    allow_nulls : boolean, default False
        Either return the intersection of all model data (False) or the union (True)
    company_info : boolean, default True
        Include additional meta data calls
    file_type : str, default "parquet"
        {"csv":csv from MDS, "parquet":parquet from MDS}
    """

    def __init__(
        self,
        ticker_list: List[str],
        config: Config = None,
        extract_drivers: bool = True,
        allow_nulls: bool = True,
        company_info: bool = True,
        file_type: str = "parquet",
    ):
        """
        Generate a dictionary of Model objects, provide ease of use functions across the dictionary.

        Parameters
        ----------
        ticker_list : list
        allow_nulls : boolean default True
            Either return the intersection of all model data (False) or the union (True)
        company_info : boolean default True
            Include additional meta data calls
        file_type : str, default "parquet"
            {"csv":csv from MDS, "parquet":parquet from MDS}
        """
        self.file_type = file_type
        self.allow_nulls = allow_nulls
        self.config = config or create_config()
        self.extract_drivers = extract_drivers
        self.log = LogFile()

        if type(ticker_list) is not list:
            if type(ticker_list) is str:
                try:
                    self.ticker_list = ticker_list.split(",")
                except:
                    self.ticker_list = [self.ticker_list]
        else:
            self.ticker_list = ticker_list

        self.drivers: DataFrame = None
        self.models: Dict[str, "Model"] = {}
        self.company_info = company_info
        self.api_headers: Dict[str, str] = get_api_headers(self.config.canalyst_api_key)

        self.get_featurelibrary()  # set self._features

    def help(self, function_name=""):
        dict_help = {
            "create_model_map": "Create a model map.  params: ticker, col_for_labels = 'time_series_name', time_series_name = 'MO_RIS_REV', tree = True, notebook = True",
            "create_time_series_chart": "Create a time series chart.  params: ticker, time_series_name",
            "guidance": "Return guidance from a mmodel.  params: ticker",
            "mrq": "Return most recent quarter from a mmodel.  params: ticker",
            "time_series_search": "Return time series regex match. params: time_series_name",
            "driver_search": "Return driver regex match. params: driver_name",
            "model_frame": "Return a DataFrame of the full modelset. params: time_series_name,period_name,is_driver='',pivot=False,mrq=False,period_duration_type='',is_historical='',n_periods='',mrq_notation=False",
            "forecast_frame": "Return a params DataFrame for use in the fit function. params: time_series_name, n_periods, function_name='value', function_value='' where function name can be add, subtract, multiply, divide, or value",
            "fit": "Return a return series for a fitted model. params: params DataFrame, return_series.  params DataFrame columns are: ticker period time_series_name value new_value",
            "model_set.models[Bloomberg Ticker].get_most_recent_model_date()": "Return most recent model upload date for a ticker",
        }
        if function_name != "":
            return dict_help[function_name]
        else:
            df = pd.DataFrame(dict_help, index=[0]).T
            df.columns = ["help"]
            return df

    def unstacked_frame(
        self, ticker="", time_series_name="", period_duration_type="", is_historical=""
    ):
        DFunc = DataFunctions()
        df = self.model_frame(
            period_duration_type=period_duration_type,
            is_historical=is_historical,
        )
        df = DFunc.unstack(df, ticker, time_series_name, "period_name_sorted")

        return df

    def time_series_formula(
        self,
        arguments=[],
        modifier="",
        time_series_name="",
        time_series_description="",
    ):
        df = self.model_frame(warning=False)
        DFunc = DataFunctions()
        df = DFunc.time_series_function(
            df,
            arguments,
            modifier,
            time_series_name,
            time_series_description,
        )
        self._features = df  # modify in place
        return

    def plot_time_series(
        self,
        ticker="",
        time_series_name="",
        mrq_notation=False,
        period_duration_type="fiscal_quarter",
        axis_labels=[["Periods", "Value"]],
    ):
        df = self.model_frame(
            ticker=ticker,
            time_series_name=time_series_name,
            period_duration_type=period_duration_type,
            is_historical=True,
        )

        df = (
            df.groupby(["ticker", "period_name_sorted"])
            .first()
            .reset_index()
            .sort_values(["ticker", "period_name_sorted"], ascending=False)
        )
        df = df[["ticker", "period_name_sorted", "value"]]
        if mrq_notation == True:
            df["Period"] = df.groupby(["ticker"]).cumcount() + 1
            df["Period"] = len(df["Period"]) - df["Period"]
            df["Period"] = df["Period"] - len(df["Period"]) - 1
        else:
            df["Period"] = df["period_name_sorted"]

        df = df.pivot(values="value", index="Period", columns="ticker")
        df.dropna()
        df_plot = df.dropna().reset_index()

        if ticker == "":
            ticker_list = self.ticker_list
        else:
            ticker_list = [ticker]

        model_chart = Chart(
            x_value=df_plot["Period"],
            y_values=df_plot[ticker_list],
            labels=ticker_list,
            title=time_series_name,
            axis_labels=axis_labels,
        )
        model_chart.show()
        return

    def common_time_series_names(self):
        """
        Return a list of time series names common to all models in the ModelSet
        """
        return self._common_time_series_names

    def plot_guidance(self, ticker, time_series_name):
        if ticker not in self.ticker_list:
            print("Please choose a ticker in this ModelSet")
            return
        self.models[ticker].plot_guidance(time_series_name)
        return

    def pe_dataset(self, ticker, yahoo_ticker, index_ticker="^GSPC", n_periods=24):
        """
        Create a join between Yahoo price data and the base _features DataFrame of the modelset.

        This function returns long form data outer joined with price data.
        Also creates a rolling beta calculation against the index input.

        Parameters
        ----------
        ticker : str
            Canalyst ticker
        yahoo_ticker : str
            Yahoo ticker
        index_ticker : str, default ^GSPC
            Index to use for betas
        n_periods : int, default 24
            number of canalyst periods

        Returns
        -------
        DataFrame

        """

        import canalyst_candas.candas_datareader as cdr

        df_earnings = cdr.get_earnings_and_prices(yahoo_ticker, index_ticker)
        if df_earnings is None:
            return None

        df_earnings = df_earnings[
            [
                "ticker",
                "earnings_date",
                "beta_252",
                "alpha_1_day",
                "alpha_5_day",
                "alpha_10_day",
            ]
        ]
        df_earnings = df_earnings.sort_values("earnings_date", ascending=False)

        df_model = self.model_frame(
            ticker=ticker, is_historical=True, period_duration_type="fiscal_quarter"
        )

        edq_list = list(
            cdr.calendar_quarter(df_earnings, "earnings_date", datetime=True)[
                "earnings_date_CALENDAR_QUARTER"
            ]
        )

        mdq_list = list(
            df_model.groupby("period_name_sorted")
            .first()
            .reset_index()
            .sort_values("period_end_date", ascending=False)["period_name_sorted"]
        )

        if len(edq_list) < n_periods:
            n_periods = len(edq_list)
            print(
                "Price earnings dataset using " + str(n_periods) + " periods available"
            )

        d = {
            "earnings_date_q": edq_list[0:n_periods],
            "period_name_sorted": mdq_list[0:n_periods],
        }

        df_dates = pd.DataFrame(d)
        df_model = pd.merge(df_model, df_dates)
        df_earnings["earnings_date_q"] = cdr.calendar_quarter(
            df_earnings, "earnings_date", datetime=True
        )["earnings_date_CALENDAR_QUARTER"]
        df_model["price_ticker"] = yahoo_ticker
        df_model = df_model.drop(columns=["ticker"])
        df_data = pd.merge(
            df_model,
            df_earnings,
            how="inner",
            left_on=["price_ticker", "earnings_date_q"],
            right_on=["ticker", "earnings_date_q"],
        )
        df_data["ticker"] = ticker
        df_data = df_data.drop(
            columns=[
                "earnings_dateshift",
                "earnings_date_CALENDAR_QUARTER",
                "category_type_slug",
                "time_series_slug",
            ]
        )
        return df_data

    def regress_dataframe_time_series_groups(
        self, df_data=None, y_name="alpha_10_day", return_grouped=True
    ):
        """
        Create grouped regressions across time series groups.

        For use with datasets created with pe_dataset function above.

        Parameters
        ----------
        df_data : DataFrame
            Dataset created with pe_dataset
        y_name : str, default "alpha_10_day"
            window for alpha (outperformance vs market) calculation
            { "alpha_10_day", "alpha_5_day", "alpha_1_day" }
        return_grouped : boolean, default True
            grouped is the default, non-grouped only for debug

        Returns
        -------
        DataFrame

        """

        ticker = df_data.iloc[0]["ticker"]
        import canalyst_candas.candas_datareader as cdr

        df = cdr.regress_dataframe_groups(
            df_data, y_name=y_name, return_grouped=return_grouped
        )
        df = df.dropna()
        df["ticker"] = ticker
        return df

    def create_model_map(
        self,
        ticker,
        col_for_labels="time_series_name",
        time_series_name="MO_RIS_REV",
        tree=True,
        notebook=False,
    ):
        """
        Return a ModelMap for a given time series name.

        Use show() to display the ModelMap.

        Parameters
        ----------
        col_for_labels : str, default time_series_name
            Label to be used on the map.
            { "time_series_name", "time_series_description" }
        time_series_name : str, default MO_RIS_REV
            String of time_series_name to be used for the ROOT node of the tree
        tree : boolean, default True
            Very complex trees might be rendered better with tree=False
        notebook : boolean, default False
            Launches tree in a separate browser window (False) or in the notebook (True)

        Returns
        -------
        ModelMap

        """
        if type(ticker) == list:
            print("Please request one ticker at a time.")
            return

        if ticker in self.ticker_list:
            model = self.models[ticker]
            return model.create_model_map(
                time_series_name=time_series_name,
                tree=tree,
                col_for_labels=col_for_labels,
                notebook=notebook,
                common_time_series_names=self._common_time_series_names,
            )
        else:
            print("Please choose a ticker in this ModelSet's ticker list")
            return

    def create_time_series_chart(
        self, ticker_list="", time_series_name="", historical=False
    ):  # MODELSET
        """
        This function will be replaced by Chart in the ModelFrame class
        """

        if ticker_list == "":
            df = self._features
            if type(time_series_name) == list:
                df = df.loc[self._features["time_series_name"].isin(time_series_name)]
            else:
                df = df.loc[self._features["time_series_name"] == time_series_name]

            if historical:
                df = df.loc[df["is_historical"] == historical]

            df = df.loc[df["period_name"].str.contains("Q")].sort_values(
                ["ticker", "period_end_date"]
            )
            df = df.loc[df["period_end_date"] > "2015-01-01"]
            df = df.dropna(subset=["value"])
            # return df
            # df = df.groupby('ticker').size().order(ascending=False).reset_index()

            row1 = df.iloc[0]
            # title = self.ticker + " " + str(row1["time_series_description"])
            xlabel = "Fiscal Quarter"
            # ylabel = row1["time_series_description"]

            colors = {
                "red": "#ff207c",
                "grey": "#C3C2C3",
                "blue": "#00838F",
                "orange": "#ffa320",
                "green": "#00ec8b",
            }

            plt.rc("figure", figsize=(12, 9))
            if type(time_series_name) != list:

                df_plot = df.loc[
                    df["time_series_name"] == time_series_name
                ].sort_values(["ticker", "time_series_name"])

                if ticker_list == "":
                    ticker_list = self.ticker_list
                elif type(ticker_list) == list:
                    ticker_list = ticker_list
                else:
                    ticker_list = [ticker_list]

                for ticker in ticker_list:

                    df_plot_one = df_plot[df_plot["ticker"] == ticker]
                    df_plot_one = df_plot_one.sort_values("period_end_date")
                    df_plot_one = df_plot_one.loc[df_plot_one["value"].notna()]

                    if df_plot_one.shape[0] > 0:
                        plt.plot(
                            df_plot_one["period_end_date"],
                            df_plot_one["value"],
                            label=ticker + " " + time_series_name,
                            linewidth=2,
                        )
                        plt.axvline(
                            x=self.models[ticker].mrq()["period_end_date"][0],
                            color="black",
                        )
            else:
                if ticker_list == "":
                    ticker_list = self.ticker_list
                elif type(ticker_list) == list:
                    ticker_list = ticker_list
                else:
                    ticker_list = [ticker_list]

                for tsn in time_series_name:
                    df_plot = df.loc[df["time_series_name"] == tsn]
                    for ticker in ticker_list:
                        df_plot_one = df_plot[df_plot["ticker"] == ticker]
                        df_plot_one = df_plot_one.sort_values("period_end_date")
                        df_plot_one = df_plot_one.loc[df_plot_one["value"].notna()]
                        if df_plot_one.shape[0] > 0:
                            plt.plot(
                                df_plot_one["period_end_date"],
                                df_plot_one["value"],
                                label=ticker + " " + tsn,
                                linewidth=2,
                            )
                            plt.axvline(
                                x=self.models[ticker].mrq()["period_end_date"][0],
                                color="black",
                            )

            plt.xticks(rotation=90)
            plt.legend()
            plt.show()
            return
        if ticker_list == "":
            ticker_list = self.ticker_list
        elif type(ticker_list) == list:
            ticker_list = ticker_list
        else:
            ticker_list = [ticker_list]

        if time_series_name == "":
            print("Please enter one time_series_name")
            return
        for ticker in ticker_list:
            df = self._features
            if historical:
                df = df.loc[df["is_historical"] == historical]

            df_plot = df.loc[df["time_series_name"] == time_series_name]
            df_plot_one = df_plot[df_plot["ticker"] == ticker]
            df_plot_one = df_plot_one.sort_values("period_end_date")
            df_plot_one = df_plot_one.loc[df_plot_one["value"].notna()]
            if df_plot_one.shape[0] > 0:
                plt.plot(
                    df_plot_one["period_end_date"],
                    df_plot_one["value"],
                    label=ticker + " " + str(time_series_name),
                    linewidth=2,
                )
                plt.axvline(
                    x=self.models[ticker].mrq()["period_end_date"][0],
                    color="black",
                )

        plt.xticks(rotation=90)
        plt.legend()
        plt.show()

        # if ticker in self.ticker_list:
        #    self.models[ticker].create_time_series_chart(time_series_name, historical)
        # else:
        #    print("Please choose a ticker in this ModelSet's ticker list")
        return

    def guidance(self, tickers: List[str], time_series_name="") -> Optional[DataFrame]:
        """
        Return a DataFrame of guidance for a given ticker(s).

        Renames the default guidance col headers with time_series_name and time_series_description for consistency.

        Parameters
        ----------
        tickers : list[str]
            A ticker in the ModelSet ticker list
        time_series_name : str, default ""
            Allow for str.contains() filtering of time series name

        Returns
        -------
        DataFrame

        """
        if self.config.s3_access_key_id == "" or self.config.s3_secret_key == "":
            print("Missing S3 keys, please contact Canalyst")

        list_df = []
        for ticker in tickers:
            if ticker in self.ticker_list:
                df = self.models[ticker].guidance()
                df["ticker"] = ticker
                col_name = "ticker"
                first_col = df.pop(col_name)
                df.insert(0, col_name, first_col)
                list_df.append(df)
            else:
                print(
                    "Please enter a list or choose a ticker in this ModelSet's ticker list"
                )
        df = pd.concat(list_df)
        df = df.rename(
            columns={"Item": "time_series_description", "Item Name": "time_series_name"}
        )
        if time_series_name != "":
            df = df.loc[
                df["time_series_description"].str.contains(time_series_name, case=False)
            ]
        return df

    def mrq(self, tickers: List[str]) -> Optional[DataFrame]:
        """
        Return the most recent quarter for a list of tickers
        """
        list_df = []
        for ticker in tickers:
            if ticker in self.ticker_list:
                df = self.models[ticker].mrq()
                df["ticker"] = ticker
                list_df.append(df)
            else:
                print("Please choose a ticker in this ModelSet's ticker list")
        return pd.concat(list_df)

    def get_featurelibrary(self, col="time_series_name"):
        """
        Creates the default DataFrame features and drivers.
        """
        list_all = []
        list_df = []

        for ticker in self.ticker_list:
            list_rows = []
            try:
                md = Model(
                    config=self.config,
                    ticker=ticker,
                    company_info=self.company_info,
                    file_type=self.file_type,
                )
                if hasattr(md, "_model_frame"):
                    self.models[ticker] = md

                    df = md._model_frame

                    list_df.append(df)

                    if df is not None:
                        list_rows = list(set(list(df[col])))
                        list_all.append(list_rows)
            except Exception as ex:
                print(f"Model error: {ticker}. Error: {ex}")

        if len(list_df) == 0:
            return

        from itertools import chain

        # faster than concat
        def fast_flatten(input_list):
            return list(chain.from_iterable(input_list))

        COLUMN_NAMES = list_df[0].columns
        df_dict: Dict[Any, list] = dict.fromkeys(COLUMN_NAMES, [])
        for col in COLUMN_NAMES:
            extracted = (df[col] for df in list_df)
            # Flatten and save to df_dict
            df_dict[col] = fast_flatten(extracted)

        df = pd.DataFrame.from_dict(df_dict)[COLUMN_NAMES]
        list_df = []
        df_dict = {}

        res = list(reduce(lambda i, j: i & j, (set(x) for x in list_all)))  # type: ignore
        self._common_time_series_names: list = res

        if self.allow_nulls == False:
            df = df_filter(df, "time_series_name", res)
            self._features = df
        else:
            self._features = df

            if df is not None:
                list_rows = list(set(list(df[col])))
                list_all.append(list_rows)
        return

    # allows for filtering and shaping the default DataFrame
    # MODELSET model_frame
    def model_frame(
        self,
        time_series_name="",
        period_name="",
        is_driver="",
        pivot=False,
        mrq=False,
        period_duration_type="",
        is_historical="",
        n_periods="",
        mrq_notation=False,
        unit_type="",
        category="",
        warning=True,
        ticker="",
    ):
        """
        Create a Pandas DataFrame with domain-specific filters.

        This function returns long form Pandas DataFrame filtered by input.

        Parameters
        ----------
        time_series_name : str or list[str], default ""
            pass a list for an exact match or a string for a regex match
        period_name : str
            a specific period name EG Q1-2022
        is_driver : boolean, default ""
            return only drivers if True or non-drivers if False, both if ""
        pivot : boolean, default ""
            pivot the DataFrame from long to wide
        mrq : boolean, default False
            return only the most recent quarter
        period_duration_type : str, default ""
            fiscal_year or fiscal_quarter
        is_historical : boolean, default ""
            only historical data if True, only forecast if False, both if ""
        n_periods : int, default ""
            number of fiscal year and or fiscal quarters to return
            use positive integers for historical periods
            use negative integers to return mrq+n periods in the case where is_historical=False
        mrq_notation : boolean, default False
            applies to pivot and applies most recent quarter / fiscal year notation
            MRQ1 MRQ2 FY1 FY2
        unit_type : str, default ""
            ratio, etc, "" returns all
        category : str, default ""
            Balance Sheet, etc, "" returns all
        warning : boolean, default True
            warns if fiscal_year or fiscal_quarter is not set
        ticker : str or list[str], default ""
            filters to a specific ticker or list of tickers, all if ""

        Returns
        -------
        DataFrame

        """

        if ticker != "":
            df = self.models[ticker]._model_frame
        else:
            df = self._features

        return filter_dataset(
            df,  # this is the core dataset of the modelset class
            time_series_name,
            period_name,
            is_driver,
            pivot,
            mrq,
            period_duration_type,
            is_historical,
            n_periods,
            mrq_notation,
            unit_type,
            category,
            warning,
        )

    def forecast_frame(
        self, time_series_name, n_periods, function_name="value", function_value=""
    ):
        """
        Builds a param DataFrame for use in the fit() function.

        Parameters
        ----------
        n_periods : int
            Number of periods to return.
        function_name : str default value
            { "multiply", "add", "divide", "subtract", "value" }
        function_value : str
            the number by which you want to apply the function or just the value to assign
            Example: if function_name = multiply then 1.1 multiplies by 10pct
            Example: if function_name = value then 5 assigns the value 5

        Returns
        ------
        DataFrame
        """

        df = self.model_frame(
            [time_series_name],
            period_duration_type="fiscal_quarter",
            is_historical=False,
            n_periods=n_periods,
        )

        if function_name != "value":
            d = FuncMusic()
            df["new_value"] = df.apply(
                lambda row: d.apply_function(
                    row["value"], modifier=function_name, argument=function_value
                ),
                axis=1,
            )
        else:
            df["new_value"] = function_value

        df = df[
            [
                "ticker",
                "period_name",
                "time_series_name",
                "time_series_description",
                "value",
                "new_value",
                "period_end_date",
            ]
        ]

        return df

    # MODELSET fit()
    def fit(self, params, return_series):
        """
        Runs the scenario engine against a param DataFrame created by forecast frame above.

        Or a user generated param DataFrame with the columns:
        ticker | period_name | time_series_name | time_series_description | value new_value | period_end_date

        Parameters
        ----------
        params : DataFrame
            Described above
        return series : str
            series to return. "MO_RIS_REV" returns revenue

        Returns
        ------
        DataFrame
        """

        # ticker period time_series_name value new_value
        dict_summary = {}
        if type(params) is dict:
            df = pd.DataFrame.from_dict(params, orient="index").reset_index()
        else:
            df = params

        df_grouped = df.groupby("ticker").first().reset_index()

        dict_data = {}
        for _, row in df_grouped.iterrows():

            df_ticker = df[df["ticker"] == row["ticker"]]
            ticker = row["ticker"]  # .iloc[0]
            self.models[ticker].set_new_uuid()
            list_changes = []

            for _, row in df_ticker.iterrows():
                feature_value = row["new_value"]
                feature_name = row["time_series_name"]
                feature_period = row["period_name"]
                ticker = row["ticker"]

                list_changes.append(
                    {
                        "time_series": feature_name,
                        "period": feature_period,
                        "value_expression": {"type": "literal", "value": feature_value},
                    }
                )

                data = {"changes": list_changes, "name": self.models[ticker].uuid}
                dict_data[ticker] = data

        ticker_list = list(set(list(df["ticker"])))

        for ticker in ticker_list:
            res = send_scenario(
                ticker,
                dict_data[ticker],
                self.api_headers,
                self.log,
                self.models[ticker].csin,
                self.models[ticker].latest_version,
                self.config.mds_host,
                self.config.verify_ssl,
            )
            try:
                self.models[ticker].record_scenario_url(res)
            except ScenarioException as e:
                error_message = {}
                print("Error code : ", e, "error")
                error_message = res.json()
                print("\n", error_message)
                continue

            try:
                self.models[ticker].model_fit(return_series)
                dict_summary[ticker] = self.models[ticker].summary()
            except Exception as e:
                print("Error on fit for " + return_series)
                print(f"Error: {e}")

        return dict_summary

    def filter_summary(self, dict_summary, period_type="Q"):

        pd.set_option("mode.chained_assignment", None)
        pd.set_option("display.float_format", lambda x: "%.5f" % x)

        list_out = []
        for ticker in dict_summary.keys():

            df = dict_summary[ticker]

            list_out.append(df)
        df = pd.concat(list_out)
        df["sort_period_name"] = (
            df["period_name"].str.split("-").str[1]
            + df["period_name"].str.split("-").str[0]
        )
        df = df.sort_values(["ticker", "sort_period_name"])
        df = df[df["period_name"].str.contains(period_type)]
        df = df.drop(columns="sort_period_name")

        return df

    def chart(
        self,
        time_series_name: Union[str, List[str]] = "",
        ticker: Union[str, List[str]] = "",
        period_duration_type: str = "fiscal_quarter",
    ) -> None:
        """
        Charts a time series with a ticker

        Usage: Can chart single time series for a single ticker: chart(time_series_name="MO_MA_Fuel", ticker="MESA US")
           or a single time series for multiple tickers: chart(time_series_name="MO_MA_Fuel", ticker=['AZUL US','MESA US'])
           or multiple time series for a single ticker: chart(time_series_name=["MO_MA_Fuel", "MO_RIS_REV"], ticker='MESA US')

        Period duration type is used to filter data by fiscal quarter or by fiscal year.

        Parameters
        ----------
        time_series_name: str or List[str]
        ticker: str or List[str]
        period_duration_type: str
        """

        def _check_for_error_handling(
            ticker: Union[str, List[str]],
            time_series_name: Union[str, List[str]],
            period_duration_type: str,
        ):
            """
            Basic error handling to check that the correct types are passed into the function
            """
            if type(ticker) == list and type(time_series_name) == list:
                raise TypeError(
                    "Ticker and time_series_name cannot both be lists. Please change one of the parameters to a single string value."
                )

            if period_duration_type not in ["fiscal_quarter", "fiscal_year"]:
                raise ValueError(
                    "Please specify fiscal_year or fiscal_quarter for period_duration_type"
                )

        def _check_for_time_series_name(
            df: DataFrame, time_series_name: Union[str, List[str]]
        ):
            """
            Checks that the time series is in the ModelSet
            """
            if type(time_series_name) == str:
                time_series_check = df[df["time_series_name"] == time_series_name]
                if time_series_check.empty is True:
                    raise ValueError(
                        f"No time series found by the name {time_series_name}, please try another time series name"
                    )
            if type(time_series_name) == list:
                for item in time_series_name:
                    if self._features[
                        self._features["time_series_name"].isin([item])
                    ].empty:
                        raise ValueError(
                            f"No time series found by the name {item}, please try another time series name"
                        )

        def _check_for_ticker(ticker_list: List[str], ticker: Union[str, List[str]]):
            """
            Checks that the ticker is in the ModelSet
            """
            if type(ticker) == str and ticker not in ticker_list:
                raise ValueError(
                    f"Ticker {ticker} not found in ModelSet. Please use a ticker from the ModelSet"
                )
            if type(ticker) == list:
                for item in ticker:
                    if item not in self.ticker_list:
                        raise ValueError(
                            f"Ticker {item} not found in ModelSet. Please use a ticker from the ModelSet"
                        )

        def _sort_chronological(df: DataFrame, period_duration_type: str):
            """
            Returns a chronoglically sorted dataframe for fiscal quarters

            Ex. Ordering of these values Q1-2021, Q2-2018 would return Q2-2018, Q1-2021
            """
            if period_duration_type == "fiscal_quarter":
                df["year"] = df.period_name.str[3:]
                df = df.sort_values(["year", "period_name"])
            return df

        _check_for_error_handling(ticker, time_series_name, period_duration_type)
        _check_for_ticker(self.ticker_list, ticker)
        _check_for_time_series_name(self._features, time_series_name)

        if type(ticker) == str and type(time_series_name) == str:
            df_plot = self._features[
                (self._features.ticker == ticker)
                & (self._features.time_series_name == time_series_name)
                & (self._features.period_duration_type == period_duration_type)
            ]

            df_plot = (
                df_plot[["time_series_name", "period_name", "value"]]
                .pivot_table(
                    values="value", index=["period_name"], columns=["time_series_name"]
                )
                .reset_index()
            )
            df_plot = _sort_chronological(df_plot, period_duration_type)

            chart = Chart(
                x_value=df_plot["period_name"],
                y_values=df_plot[time_series_name],
                labels=[time_series_name],
                axis_labels=[["Periods", "Value"]],
                title=time_series_name,
            )
            chart.show()

        elif type(ticker) == list and type(time_series_name) == str:
            df_plot = self._features[
                (self._features.period_duration_type == period_duration_type)
                & (self._features.time_series_name == time_series_name)
            ]

            df_plot = (
                df_plot[df_plot["ticker"].isin(ticker)][
                    ["ticker", "period_name", "value"]
                ]
                .pivot_table(values="value", index=["period_name"], columns=["ticker"])
                .reset_index()
            )
            df_plot = _sort_chronological(df_plot, period_duration_type)

            chart = Chart(
                x_value=df_plot["period_name"],
                y_values=df_plot[ticker],
                labels=ticker,
                axis_labels=[["Periods", "Value"]],
                title=time_series_name,
            )

            chart.show()

        elif type(time_series_name) == list and type(ticker) == str:
            df_plot = self._features[
                (self._features.period_duration_type == period_duration_type)
                & self._features["time_series_name"].isin(time_series_name)
            ]

            df_plot = (
                df_plot[df_plot["ticker"].isin([ticker])][
                    ["time_series_name", "period_name", "value"]
                ]
                .pivot_table(
                    values="value", index=["period_name"], columns=["time_series_name"]
                )
                .reset_index()
            )
            df_plot = _sort_chronological(df_plot, period_duration_type)

            chart = Chart(
                x_value=df_plot["period_name"],
                y_values=df_plot[time_series_name],
                labels=time_series_name,
                axis_labels=[["Periods", "Value"]],
                title=ticker,
            )
            chart.show()
        else:
            raise TypeError(
                "Incorrect parameter types. Please refer to the function documentation for proper usage of the function"
            )
